#!/usr/bin/env python3
"""Sparkle CLI entry point."""

import sys
import os
from pathlib import Path

module_path = Path(__file__).parent.resolve()

package_cli_entry_points = [
    module_path.parent / "solver" / "solver_cli.py",
    module_path.parent / "selector" / "selector_cli.py",
    module_path.parent / "selector" / "extractor_cli.py",
    module_path.parent / "configurator" / "configurator_cli.py",
]


def commands() -> list[str]:
    """Get list of available commands."""
    module_path = Path(__file__).parent.resolve()
    self_name = Path(__file__).name
    return [
        path.stem
        for path in module_path.iterdir()
        if path.is_file()
        and path.suffix == ".py"
        and path.name != self_name
        and path.name != "__init__.py"
    ]


def main() -> None:
    """Pass through command to launch CLI commands."""
    max_space = max(
        [path.name.count("_") for path in module_path.iterdir() if path.is_file()]
    )
    if len(sys.argv) < 2:
        print("Usage: sparkle <command>")
        sys.exit(1)
    if " ".join(sys.argv[1:]) == "install autocomplete":
        import urllib.request

        print(sys.prefix)
        if sys.prefix == sys.base_prefix:
            print(
                "Sparkle is not installed in a virtual environment! "
                "Autocomplete must be installed manually, see the documentation."
            )
            sys.exit(-1)
        # TODO: Update this URL to link to the file on main
        code_inject = (
            urllib.request.urlopen(
                "https://raw.githubusercontent.com/ADA-research/Sparkle/refs/heads/development/Resources/Other/venv_autocomplete.sh"
            )
            .read()
            .decode()
        )
        venv_profile_path = Path(sys.prefix) / "bin" / "activate"
        if venv_profile_path.is_file():
            if code_inject in venv_profile_path.read_text():
                print(
                    f"[{Path(sys.prefix).name}] Sparkle autocomplete is already installed in the virtual environment! Exit."
                )
                sys.exit(-1)
            with venv_profile_path.open("a") as f:
                f.write(code_inject)
            print(
                f"[{Path(sys.prefix).name}] Sparkle autocomplete has been installed in the virtual environment!"
            )
            sys.exit(0)
        print(
            "Virtual environment not found! Manual installation in activation script required, see the documentation."
        )
        sys.exit(-1)
    # Support spaces instead of _
    possible_commands = commands()
    command, command_file = "", Path()
    for i in range(1, min(max_space, len(sys.argv))):
        if "--" in sys.argv[i]:  # Parameter is never part of the command
            break
        command = "_".join(sys.argv[1 : i + 1])
        args = sys.argv[i + 1 :]
        command_file = module_path / f"{command}.py"
        if command in possible_commands:
            break

    if command_file.is_file():
        os.system(f"python3 {command_file} {' '.join(args)}")
        sys.exit(0)
    else:
        print(f"Sparkle does not understand command <{command}>", end="")
        from difflib import SequenceMatcher

        similarities = [
            SequenceMatcher(None, command, alt).ratio() for alt in possible_commands
        ]

        if max(similarities) > 0.6:
            alternative = possible_commands[similarities.index(max(similarities))]
            print(f". Did you mean <{alternative}>?")
            sys.exit(-2)

        sys.exit(-1)


if __name__ == "__main__":
    main()



"""Objects and methods relating to instances for Sparkle."""

from __future__ import annotations
from pathlib import Path

import csv
import numpy as np


class InstanceSet:
    """Base object representation of a set of instances."""

    def __init__(self: InstanceSet, target: Path | list[str, Path]) -> None:
        """Initialise an Instances object from a directory.

        Args:
            target: The Path, or list of paths to create the instance set from.
        """
        self.directory: Path = target
        self._instance_names: list[str] = []
        self._instance_paths: list[Path] = []

    @property
    def size(self: InstanceSet) -> int:
        """Returns the number of instances in the set."""
        return len(self._instance_paths)

    @property
    def all_paths(self: InstanceSet) -> list[Path]:
        """Returns all file paths in the instance set as a flat list."""
        return self._instance_paths

    @property
    def instance_paths(self: InstanceSet) -> list[Path]:
        """Get processed instance paths."""
        return self._instance_paths

    @property
    def instance_names(self: InstanceSet) -> list[str]:
        """Get processed instance names for instances."""
        return self._instance_names

    @property
    def instances(self: InstanceSet) -> list[str]:
        """Get instance names with relative path."""
        return [str(p.with_suffix("")) for p in self._instance_paths]

    @property
    def name(self: InstanceSet) -> str:
        """Get instance set name."""
        return self.directory.name

    def __str__(self: InstanceSet) -> str:
        """Get the string representation of an Instance Set."""
        return self.name

    def __repr__(self: InstanceSet) -> str:
        """Get detailed representation of an Instance Set."""
        return (
            f"{self.name}:\n"
            f"\t- Type: {type(self).__name__}\n"
            f"\t- Directory: {self.directory}\n"
            f"\t- # Instances: {self.size}"
        )

    def get_path_by_name(self: InstanceSet, name: str) -> Path | list[Path]:
        """Retrieves an instance paths by its name. Returns None upon failure."""
        for idx, instance_name in enumerate(self._instance_names):
            if instance_name == name:
                return self._instance_paths[idx]
        return None


class FileInstanceSet(InstanceSet):
    """Object representation of a set of single-file instances."""

    def __init__(self: FileInstanceSet, target: Path) -> None:
        """Initialise an InstanceSet, where each instance is a file in the directory.

        Args:
            target: Path to the instances directory. If multiple files are found,
                they are assumed to have the same number of instances per file.
        """
        super().__init__(target)
        self._name: str = target.stem
        if target.is_file():
            # Single instance set
            self._instance_paths = [target]
            self._instance_names = [target.stem]
            self.directory = target.parent
        else:
            # Default situation, treat each file in the directory as an instance
            self._instance_paths = [p for p in self.directory.iterdir()]
            self._instance_names = [p.stem for p in self._instance_paths]

    @property
    def name(self: FileInstanceSet) -> str:
        """Get instance set name."""
        return self._name


class MultiFileInstanceSet(InstanceSet):
    """Object representation of a set of multi-file instances."""

    instance_csv = "instances.csv"

    def __init__(self: MultiFileInstanceSet, target: Path) -> None:
        """Initialise an Instances object from a directory.

        Args:
            target: Path to the instances directory. Will read from instances.csv.
        """
        target_dir = target.parent if not target.is_dir() else target
        super().__init__(target_dir)
        # A path pointing to the directory of instances
        self.instance_file = self.directory / MultiFileInstanceSet.instance_csv
        # Read from instance_file
        if not target.is_dir():
            # Single file
            instance_list = [
                line
                for line in csv.reader(self.instance_file.open())
                if target.stem in line
            ]
        else:
            # Multi file
            instance_list = [line for line in csv.reader(self.instance_file.open())]

        for instance in instance_list:
            self._instance_names.append(instance[0])
            self._instance_paths.append(
                [(self.directory / f) if isinstance(f, str) else f for f in instance[1:]]
            )

    @property
    def all_paths(self: MultiFileInstanceSet) -> list[Path]:
        """Returns all file paths in the instance set as a flat list."""
        return [p for instance in self._instance_paths for p in instance] + [
            self.instance_file
        ]

    @property
    def instances(self: InstanceSet) -> list[str]:
        """Get instance names with relative path for multi-file instances."""
        return [self.directory / inst_name for inst_name in self.instance_names]


class IterableFileInstanceSet(InstanceSet):
    """Object representation of files containing multiple instances."""

    supported_filetypes = set([".csv", ".npy"])

    def __init__(self: IterableFileInstanceSet, target: Path) -> None:
        """Initialise an InstanceSet from a single file.

        Args:
            target: Path to the instances directory. If multiple files are found,
                they are assumed to have the same number of instances.
        """
        super().__init__(target)
        self._instance_paths = [
            p
            for p in self.directory.iterdir()
            if p.suffix in IterableFileInstanceSet.supported_filetypes
        ]
        self._size = IterableFileInstanceSet.__determine_size__(self._instance_paths[0])
        self._instance_names = [p.name for p in self._instance_paths]

    @property
    def size(self: IterableFileInstanceSet) -> int:
        """Returns the number of instances in the set."""
        return self._size

    @staticmethod
    def __determine_size__(file: Path) -> int:
        """Determine the number of instances in a file."""
        match file.suffix:
            case ".csv":
                return len(file.open().readlines())
            case ".npy":
                return len(np.load(file))



"""Configurator classes to implement SMAC3 in Sparkle."""

from __future__ import annotations
from pathlib import Path

from smac import version as smac_version
from smac import Scenario as SmacScenario
from smac import facade as smacfacades
from smac.runhistory.enumerations import StatusType as SmacStatusType
import numpy as np
import random
from typing import Optional

from runrunner import Runner, Run

from sparkle.configurator.configurator import Configurator, ConfigurationScenario
from sparkle.solver import Solver
from sparkle.structures import FeatureDataFrame, PerformanceDataFrame
from sparkle.instance import InstanceSet, Instance_Set
from sparkle.types import SparkleObjective, resolve_objective, SolverStatus


class SMAC3(Configurator):
    """Class for SMAC3 (Python) configurator."""

    configurator_path = Path(__file__).parent.resolve() / "SMAC3"
    configurator_target = configurator_path / "smac3_target_algorithm.py"

    full_name = "Sequential Model-based Algorithm Configuration"
    version = smac_version

    def __init__(self: SMAC3) -> None:
        """Returns the SMAC3 configurator, Python SMAC V2.3.1."""
        return super().__init__(multi_objective_support=False)

    @property
    def name(self: SMAC3) -> str:
        """Returns the name of the configurator."""
        return SMAC3.__name__

    @staticmethod
    def scenario_class() -> ConfigurationScenario:
        """Returns the SMAC3 scenario class."""
        return SMAC3Scenario

    @staticmethod
    def check_requirements(verbose: bool = False) -> bool:
        """Check that SMAC3 is installed."""
        return True  # Is automatically installed with Sparkle

    @staticmethod
    def download_requirements() -> None:
        """Download SMAC3."""
        return  # Nothing to do

    def configure(
        self: SMAC3,
        scenario: SMAC3Scenario,
        data_target: PerformanceDataFrame,
        validate_after: bool = True,
        sbatch_options: list[str] = [],
        slurm_prepend: str | list[str] | Path = None,
        num_parallel_jobs: int = None,
        base_dir: Path = None,
        run_on: Runner = Runner.SLURM,
    ) -> list[Run]:
        """Start configuration job.

        Args:
            scenario: ConfigurationScenario object
            data_target: PerformanceDataFrame where to store the found configurations
            validate_after: Whether the Validator will be called after the configuration
            sbatch_options: List of slurm batch options to use
            slurm_prepend: Slurm script to prepend to the sbatch
            num_parallel_jobs: The maximum number of jobs to run parallel.
            base_dir: The path where the sbatch scripts will be created for Slurm.
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        scenario.create_scenario()
        if (
            scenario.smac3_scenario.walltime_limit
            == scenario.smac3_scenario.cputime_limit
            == np.inf
        ):
            print("WARNING: Starting SMAC3 scenario without any time limit.")
        configuration_ids = scenario.configuration_ids

        # Scenario file also has a seed, but not for all type of configurators
        seeds = [random.randint(0, 2**32 - 1) for _ in range(scenario.number_of_runs)]
        num_parallel_jobs = num_parallel_jobs or scenario.number_of_runs
        # We do not require the configurator CLI as its already our own python wrapper
        cmds = [
            f"python3 {self.configurator_target.absolute()} "
            f"{scenario.scenario_file_path.absolute()} {configuration_id} {seed} "
            f"{data_target.csv_filepath}"
            for configuration_id, seed in zip(configuration_ids, seeds)
        ]
        return super().configure(
            configuration_commands=cmds,
            data_target=data_target,
            output=None,
            scenario=scenario,
            configuration_ids=configuration_ids,
            validate_after=validate_after,
            sbatch_options=sbatch_options,
            slurm_prepend=slurm_prepend,
            num_parallel_jobs=num_parallel_jobs,
            base_dir=base_dir,
            run_on=run_on,
        )

    @staticmethod
    def organise_output(
        output_source: Path,
        output_target: Path,
        scenario: SMAC3Scenario,
        configuration_id: str,
    ) -> None | str:
        """Method to restructure and clean up after a single configurator call."""
        import json

        if not output_source.exists():
            print(f"SMAC3 ERROR: Output source file does not exist! [{output_source}]")
            return
        results_dict = json.load(output_source.open("r"))
        configurations = [value for _, value in results_dict["configs"].items()]
        config_evals = [[] for _ in range(len(configurations))]
        objective = scenario.sparkle_objective
        for entry in results_dict["data"]:
            smac_conf_id = entry["config_id"]
            score = entry["cost"]
            # SMAC3 configuration ids start at 1
            config_evals[smac_conf_id - 1].append(score)
        config_evals = [
            objective.instance_aggregator(evaluations) for evaluations in config_evals
        ]
        best_config = configurations[
            config_evals.index(objective.solver_aggregator(config_evals))
        ]
        return Configurator.save_configuration(
            scenario, configuration_id, best_config, output_target
        )

    def get_status_from_logs(self: SMAC3) -> None:
        """Method to scan the log files of the configurator for warnings."""
        raise NotImplementedError

    @staticmethod
    def convert_status(status: SolverStatus) -> SmacStatusType:
        """Converts Sparkle Solver status to SMAC3 target status."""
        mapping = {
            SolverStatus.SUCCESS: SmacStatusType.SUCCESS,
            SolverStatus.CRASHED: SmacStatusType.CRASHED,
            SolverStatus.TIMEOUT: SmacStatusType.TIMEOUT,
            SolverStatus.WRONG: SmacStatusType.CRASHED,
            SolverStatus.UNKNOWN: SmacStatusType.CRASHED,
            SolverStatus.ERROR: SmacStatusType.CRASHED,
            SolverStatus.KILLED: SmacStatusType.TIMEOUT,
            SolverStatus.SAT: SmacStatusType.SUCCESS,
            SolverStatus.UNSAT: SmacStatusType.SUCCESS,
        }
        return mapping[status]


class SMAC3Scenario(ConfigurationScenario):
    """Class to handle SMAC3 configuration scenarios."""

    def __init__(
        self: SMAC3Scenario,
        solver: Solver,
        instance_set: InstanceSet,
        sparkle_objectives: list[SparkleObjective],
        number_of_runs: int,
        parent_directory: Path,
        solver_cutoff_time: int = None,
        smac_facade: smacfacades.AbstractFacade
        | str = smacfacades.AlgorithmConfigurationFacade,
        crash_cost: float | list[float] = np.inf,
        termination_cost_threshold: float | list[float] = np.inf,
        walltime_limit: float = np.inf,
        cputime_limit: float = np.inf,
        solver_calls: int = None,
        use_default_config: bool = False,
        feature_data: FeatureDataFrame | Path = None,
        min_budget: float | int | None = None,
        max_budget: float | int | None = None,
        seed: int = -1,
        n_workers: int = 1,
        max_ratio: float = None,
        smac3_output_directory: Path = Path(),
        timestamp: str = None,
    ) -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver
                The solver to use for configuration.
            instance_set: InstanceSet
                The instance set to use for configuration.
            sparkle_objectives: list[SparkleObjective]
                The objectives to optimize.
            number_of_runs: int
                The number of times this scenario will be executed with different seeds.
            parent_directory: Path
                The parent directory where the configuration files will be stored.
            solver_cutoff_time: int
                Maximum CPU runtime in seconds that each solver call (trial)
                is allowed to run. Is managed by RunSolver, not pynisher.
            smac_facade: AbstractFacade, defaults to AlgorithmConfigurationFacade
                The SMAC facade to use for Optimisation.
            crash_cost: float | list[float], defaults to np.inf
                Defines the cost for a failed trial. In case of multi-objective,
                each objective can be associated with a different cost.
            termination_cost_threshold: float | list[float], defaults to np.inf
                Defines a cost threshold when the optimization should stop. In case of
                multi-objective, each objective *must* be associated with a cost.
                The optimization stops when all objectives crossed the threshold.
            walltime_limit: float, defaults to np.inf
                The maximum time in seconds that SMAC is allowed to run. Only counts
                solver time.
            cputime_limit: float, defaults to np.inf
                The maximum CPU time in seconds that SMAC is allowed to run. Only counts
                solver time.
            solver_calls: int, defaults to None
                The maximum number of trials (combination of configuration, seed, budget,
                and instance, depending on the task) to run. If left as None, will be
                calculated as int(cutoff time / cputime or walltime limit)
            use_default_config: bool, defaults to False
                If True, the configspace's default configuration is evaluated in the
                initial design. For historic benchmark reasons, this is False by default.
                Notice, that this will result in n_configs + 1 for the initial design.
                Respecting n_trials, this will result in one fewer evaluated
                configuration in the optimization.
            instances: list[str] | None, defaults to None
                Names of the instances to use. If None, no instances are used. Instances
                could be dataset names, seeds, subsets, etc.
            feature_data: FeatureDataFrame or Path, defaults to None
                Instances can be associated with features. For example, meta data of
                the dataset (mean, var, ...) can be incorporated which are then further
                used to expand the training data of the surrogate model. If Path, loaded
                from file. When no features are given, uses index as instance features.
            min_budget: float | int | None, defaults to None
                The minimum budget (epochs, subset size, number of instances, ...) that
                is used for the optimization. Use this argument if you use multi-fidelity
                or instance optimization.
            max_budget: float | int | None, defaults to None
                The maximum budget (epochs, subset size, number of instances, ...) that
                is used for the optimization. Use this argument if you use multi-fidelity
                or instance optimization.
            seed: int, defaults to -1
                The seed is used to make results reproducible.
                If seed is -1, SMAC will generate a random seed.
            n_workers: int, defaults to 1
                The number of workers to use for parallelization.
                If `n_workers` is greather than 1, SMAC will use DASK to parallelize the
                optimization.
            max_ratio: float, defaults to None.
                Facade uses at most scenario.n_trials * max_ratio number of
                configurations in the initial design. Additional configurations are not
                affected by this parameter. Not applicable to each facade.
            smac3_output_directory: Path, defaults to Path()
                The output subdirectory for the SMAC3 scenario. Defaults to the scenario
                results directory.
            timestamp: An optional timestamp for the directory name.
        """
        super().__init__(
            solver,
            instance_set,
            sparkle_objectives,
            number_of_runs,
            parent_directory,
            timestamp,
        )
        self.feature_data = feature_data
        if isinstance(self.feature_data, Path):  # Load from file
            self.feature_data = FeatureDataFrame(self.feature_data)

        # Facade parameters
        self.smac_facade = smac_facade
        if isinstance(self.smac_facade, str):
            self.smac_facade = getattr(smacfacades, self.smac_facade)
        self.max_ratio = max_ratio

        if self.feature_data is not None:
            instance_features = {
                instance: self.feature_data.get_instance(str(instance))
                for instance in self.instance_set.instance_paths
            }
        else:
            # 'If no instance features are passed, the runhistory encoder can not
            # distinguish between different instances and therefore returns the same data
            # points with different values, all of which are used to train the surrogate
            # model. Consider using instance indices as features.'
            instance_features = {
                name: [index] for index, name in enumerate(instance_set.instance_paths)
            }

        # NOTE: Patchfix; SMAC3 can handle MO but Sparkle also gives non-user specified
        # objectives but not all class methods can handle it here yet
        self.sparkle_objective = sparkle_objectives[0]

        # NOTE: We don't use trial_walltime_limit as a way of managing resources
        # As it uses pynisher to do it (python based) and our targets are maybe not
        # RunSolver is the better option for accuracy.
        self.solver_cutoff_time = solver_cutoff_time
        if solver_calls is None:  # If solver calls is None, try to calculate it
            if self.solver_cutoff_time is not None and (cputime_limit or walltime_limit):
                if cputime_limit:
                    solver_calls = int(cputime_limit / self.solver_cutoff_time)
                elif walltime_limit:
                    solver_calls = int(walltime_limit / self.solver_cutoff_time)
            else:
                solver_calls = 100  # SMAC3 Default value
        self.smac3_output_directory = smac3_output_directory
        self.crash_cost = crash_cost
        self.termination_cost_threshold = termination_cost_threshold
        self.walltime_limit = walltime_limit
        self.cputime_limit = cputime_limit
        self.solver_calls = solver_calls
        self.use_default_config = use_default_config
        self.instance_features = instance_features
        self.min_budget = min_budget
        self.max_budget = max_budget
        self.seed = seed
        self.n_workers = n_workers
        self.smac3_scenario: Optional[SmacScenario] = None

    def create_scenario(self: SMAC3Scenario) -> None:
        """This prepares all the necessary subdirectories related to configuration."""
        super().create_scenario()
        self.log_dir.mkdir(parents=True)
        if self.smac3_scenario is None:
            self.set_smac3_scenario()
        self.create_scenario_file()

    def set_smac3_scenario(self: SMAC3Scenario) -> None:
        """Set the smac scenario object."""
        self.smac3_scenario = SmacScenario(
            configspace=self.solver.get_configuration_space(),
            name=self.name,
            output_directory=self.results_directory / self.smac3_output_directory,
            deterministic=self.solver.deterministic,
            objectives=[self.sparkle_objective.name],
            crash_cost=self.crash_cost,
            termination_cost_threshold=self.termination_cost_threshold,
            walltime_limit=self.walltime_limit,
            cputime_limit=self.cputime_limit,
            n_trials=self.solver_calls,
            use_default_config=self.use_default_config,
            instances=self.instance_set.instance_paths,
            instance_features=self.instance_features,
            min_budget=self.min_budget,
            max_budget=self.max_budget,
            seed=self.seed,
            n_workers=self.n_workers,
        )

    @property
    def log_dir(self: SMAC3Scenario) -> Path:
        """Return the path of the log directory."""
        if self.directory:
            return self.directory / "logs"
        return None

    @property
    def configurator(self: SMAC3Scenario) -> SMAC3:
        """Return the type of configurator the scenario belongs to."""
        return SMAC3

    def create_scenario_file(self: SMAC3Scenario) -> Path:
        """Create a file with the configuration scenario."""
        with self.scenario_file_path.open("w") as file:
            for key, value in self.serialise().items():
                file.write(f"{key} = {value}\n")

    def serialise(self: SMAC3Scenario) -> dict:
        """Serialize the configuration scenario."""
        feature_data = str(self.feature_data.csv_filepath) if self.feature_data else None
        return {
            "solver": self.solver.directory,
            "instance_set": self.instance_set.directory,
            "sparkle_objectives": ",".join(self.smac3_scenario.objectives),
            "solver_cutoff_time": self.solver_cutoff_time,
            "number_of_runs": self.number_of_runs,
            "smac_facade": self.smac_facade.__name__,
            "crash_cost": self.smac3_scenario.crash_cost,
            "termination_cost_threshold": self.smac3_scenario.termination_cost_threshold,
            "walltime_limit": self.smac3_scenario.walltime_limit,
            "cputime_limit": self.smac3_scenario.cputime_limit,
            "solver_calls": self.smac3_scenario.n_trials,
            "use_default_config": self.smac3_scenario.use_default_config,
            "feature_data": feature_data,
            "min_budget": self.smac3_scenario.min_budget,
            "max_budget": self.smac3_scenario.max_budget,
            "seed": self.smac3_scenario.seed,
            "n_workers": self.smac3_scenario.n_workers,
        }

    @staticmethod
    def from_file(scenario_file: Path, run_index: int = None) -> SMAC3Scenario:
        """Reads scenario file and initalises ConfigurationScenario.

        Args:
            scenario_file: Path to scenario file.
            run_index: If given, reads as the scenario with run_index for offset
                in output directory and seed.

        Returns:
            ConfigurationScenario.
        """
        import ast

        variables = {
            keyvalue[0]: keyvalue[1].strip()
            for keyvalue in (
                line.split(" = ", maxsplit=1)
                for line in scenario_file.open().readlines()
                if line.strip() != ""
            )
        }
        variables["solver"] = Solver(Path(variables["solver"]))
        variables["instance_set"] = Instance_Set(Path(variables["instance_set"]))
        variables["sparkle_objectives"] = [
            resolve_objective(o) for o in variables["sparkle_objectives"].split(",")
        ]
        variables["parent_directory"] = scenario_file.parent.parent
        variables["solver_cutoff_time"] = int(variables["solver_cutoff_time"])
        variables["number_of_runs"] = int(variables["number_of_runs"])
        variables["smac_facade"] = getattr(smacfacades, variables["smac_facade"])

        # We need to support both lists of floats and single float (np.inf is fine)
        if variables["crash_cost"].startswith("["):
            variables["crash_cost"] = [
                float(v) for v in ast.literal_eval(variables["crash_cost"])
            ]
        else:
            variables["crash_cost"] = float(variables["crash_cost"])
        if variables["termination_cost_threshold"].startswith("["):
            variables["termination_cost_threshold"] = [
                float(v)
                for v in ast.literal_eval(variables["termination_cost_threshold"])
            ]
        else:
            variables["termination_cost_threshold"] = float(
                variables["termination_cost_threshold"]
            )

        variables["walltime_limit"] = float(variables["walltime_limit"])
        variables["cputime_limit"] = float(variables["cputime_limit"])
        variables["solver_calls"] = ast.literal_eval(variables["solver_calls"])
        variables["use_default_config"] = ast.literal_eval(
            variables["use_default_config"]
        )

        if variables["feature_data"] != "None":
            variables["feature_data"] = Path(variables["feature_data"])
        else:
            variables["feature_data"] = None

        variables["min_budget"] = ast.literal_eval(variables["min_budget"])
        variables["max_budget"] = ast.literal_eval(variables["max_budget"])

        variables["seed"] = ast.literal_eval(variables["seed"])
        variables["n_workers"] = ast.literal_eval(variables["n_workers"])
        if run_index is not None:  # Offset
            variables["seed"] += run_index
            variables["smac3_output_directory"] = Path(f"run_{run_index}")

        timestamp = scenario_file.parent.name.split("_")[-1]
        scenario = SMAC3Scenario(**variables, timestamp=timestamp)
        scenario.set_smac3_scenario()
        return scenario



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Execute Sparkle portfolio selector, read/write to SelectionScenario."""

import argparse
import sys
from filelock import FileLock, Timeout
from pathlib import Path

from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.solver import Solver
from sparkle.selector import SelectionScenario
from sparkle.instance import Instance_Set


def main(argv: list[str]) -> None:
    """Main function of the Selector CLI."""
    # Define command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--selector-scenario",
        required=True,
        type=Path,
        help="path to portfolio selector scenario",
    )
    parser.add_argument(
        "--instance", required=True, type=Path, help="path to instance to run on"
    )
    parser.add_argument(
        "--feature-data", required=True, type=Path, help="path to feature data"
    )
    parser.add_argument(
        "--seed",
        type=int,
        required=False,
        help="seed to use for the solver. If not provided, read from "
        "the PerformanceDataFrame or generate one.",
    )
    parser.add_argument(
        "--log-dir", type=Path, required=False, help="path to the log directory"
    )
    args = parser.parse_args(argv)

    # Process command line arguments
    selector_scenario = SelectionScenario.from_file(args.selector_scenario)
    feature_data = FeatureDataFrame(Path(args.feature_data))
    instance_set = Instance_Set(args.instance)
    instance = str(instance_set.instance_paths[0])
    instance_name = instance_set.instance_names[0]
    seed = args.seed
    if seed is None:
        # Try to read from PerformanceDataFrame
        seed = selector_scenario.selector_performance_data.get_value(
            selector_scenario.__selector_solver_name__,
            instance_name,
            solver_fields=[PerformanceDataFrame.column_seed],
        )
        if seed is None:  # Still no value
            import random

            seed = random.randint(0, 2**32 - 1)

    # Note: Following code could be adjusted to run entire instance set
    # Run portfolio selector
    print(f"Sparkle portfolio selector predicting for instance {instance_name} ...")
    feature_instance_name = instance_name
    if instance_name not in feature_data.instances:
        if Path(instance_name).name in feature_data.instances:
            feature_instance_name = Path(instance_name).name
        elif str(instance) in feature_data.instances:
            feature_instance_name = str(instance)
        elif str(Path(instance).with_suffix("")) in feature_data.instances:
            feature_instance_name = str(Path(instance).with_suffix(""))
        else:
            raise ValueError(
                f"Could not resolve {instance} features in {feature_data.csv_filepath}"
            )
    predict_schedule = selector_scenario.selector.run(
        selector_scenario.selector_file_path, feature_instance_name, feature_data
    )

    if predict_schedule is None:  # Selector Failed to produce prediction
        sys.exit(-1)

    print("Predicting done! Running schedule...")
    performance_data = selector_scenario.selector_performance_data
    selector_output = {}
    for solver, config_id, cutoff_time in predict_schedule:
        config = performance_data.get_full_configuration(solver, config_id)
        solver = Solver(Path(solver))
        print(
            f"\t- Calling {solver.name} ({config_id}) with time budget {cutoff_time} "
            f"on instance {instance}..."
        )
        solver_output = solver.run(  # Runs locally by default
            instance,
            objectives=[selector_scenario.objective],
            seed=seed,
            cutoff_time=cutoff_time,
            configuration=config,
            log_dir=args.log_dir,
        )
        for key in solver_output:
            if key in selector_output and isinstance(solver_output[key], (int, float)):
                selector_output[key] += solver_output[key]
            else:
                selector_output[key] = solver_output[key]
        print(f"\t- Calling solver {solver.name} ({config_id}) done!")

        if solver_output["status"].positive:
            print(f"{instance} was solved by {solver.name} ({config_id})")
            break
        print(f"{instance} is not solved in this call")

    selector_value = selector_output[selector_scenario.objective.name]
    if selector_scenario.objective.post_process:
        selector_value = selector_scenario.objective.post_process(
            selector_output[selector_scenario.objective.name],
            cutoff_time,
            selector_output["status"],
        )
    if solver_output["status"].positive:
        print(
            f"Selector {selector_scenario.selector.name} solved {instance} "
            f"with a value of {selector_value} ({selector_scenario.objective.name})."
        )
    else:
        print(f"Selector {selector_scenario.selector.name} did not solve {instance}.")
    print(f"Writing results to {performance_data.csv_filepath} ...")
    try:
        # Creating a seperate locked file for writing
        lock = FileLock(f"{performance_data.csv_filepath}.lock")
        with lock.acquire(timeout=60):
            # Reload the dataframe to latest version
            performance_data = PerformanceDataFrame(performance_data.csv_filepath)
            performance_data.set_value(
                selector_value,
                selector_scenario.__selector_solver_name__,
                instance_name,
                objective=selector_scenario.objective.name,
            )
            performance_data.save_csv()
        lock.release()
    except Timeout:
        print(f"ERROR: Cannot acquire File Lock on {performance_data}.")


if __name__ == "__main__":
    main(sys.argv[1:])



"""Target algorithm for SMAC3 to allow for Slurm scheduling."""

from __future__ import annotations
import sys
from pathlib import Path
from typing import Any, Callable
import copy
import inspect
import math
import traceback
from functools import partial

import numpy as np
from ConfigSpace import Configuration
from smac.runner.abstract_runner import StatusType
from smac.runner.abstract_serial_runner import AbstractSerialRunner
from smac.scenario import Scenario

from sparkle.configurator.implementations import SMAC3, SMAC3Scenario


def smac3_solver_call(config: Configuration, instance: str, seed: int) -> list[float]:
    """Wrapper function."""
    return solver.run(
        instance,
        objectives,
        seed,
        cutoff_time=cutoff,
        configuration=dict(config),
        log_dir=log_dir,
    )


class SparkleTargetFunctionRunner(AbstractSerialRunner):
    """Sparkle override class to execute target functions which are python functions.

    Ripped from SMAC3, modified to report RunSolver measurements instead of Python time.

    Parameters
    ----------
    target_function: Callable
        The target function. Set to default to smac3_solver_call.
    scenario: Scenario
    required_arguments: list[str], defaults to []
        A list of required arguments, which are passed to the target function.
    """

    def __init__(
        self: SparkleTargetFunctionRunner,
        scenario: Scenario,
        target_function: Callable = smac3_solver_call,
        required_arguments: list[str] = None,
    ) -> None:
        """Initialize SparkleTargetFunctionRunner."""
        if required_arguments is None:
            required_arguments = []
        super().__init__(scenario=scenario, required_arguments=required_arguments)
        self._target_function = target_function

        # Signatures here
        signature = inspect.signature(self._target_function).parameters
        for argument in required_arguments:
            if argument not in signature.keys():
                raise RuntimeError(
                    f"Target function needs to have the arguments {required_arguments} "
                    f"but could not find {argument}."
                )

        # Now we check for additional arguments which are not used by SMAC
        # However, we only want to warn the user and not
        for key in list(signature.keys())[1:]:
            if key not in required_arguments:
                print(
                    f"The argument {key} is not set by SMAC: Consider removing it "
                    "from the target function."
                )

        # Pynisher limitations
        if (memory := self._scenario.trial_memory_limit) is not None:
            unit = None
            if isinstance(memory, (tuple, list)):
                memory, unit = memory
            memory = int(math.ceil(memory))
            if unit is not None:
                memory = (memory, unit)

        if (time := self._scenario.trial_walltime_limit) is not None:
            time = int(math.ceil(time))

        self._memory_limit = memory
        self._algorithm_walltime_limit = time

    @property
    def meta(self: SparkleTargetFunctionRunner) -> dict[str, Any]:  # noqa: D102
        meta = super().meta

        # Partial's don't have a __code__ attribute but are a convenient
        # way a user might want to pass a function to SMAC, specifying
        # keyword arguments.
        f = self._target_function
        if isinstance(f, partial):
            f = f.func
            meta.update({"code": str(f.__code__.co_code)})
            meta.update({"code-partial-args": repr(f)})
        else:
            meta.update({"code": str(self._target_function.__code__.co_code)})

        return meta

    def run(
        self: SparkleTargetFunctionRunner,
        config: Configuration,
        instance: str | None = None,
        budget: float | None = None,
        seed: int | None = None,
        **dask_data_to_scatter: dict[str, Any],
    ) -> tuple[StatusType, float | list[float], float, dict]:
        """Calls the target function and parses output values for SMAC3.

        Parameters
        ----------
        config: Configuration
            Configuration to be passed to the target function.
        instance: str | None, defaults to None
            The Problem instance.
        budget: float | None, defaults to None
            A positive, real-valued number representing an arbitrary limit to the
            target function handled by the target function internally.
        seed: int | None, defaults to None
            The seed to use.
        dask_data_to_scatter: dict[str, Any]
            This kwargs must be empty when we do not use dask! ()
            When a user scatters data from their local process to the distributed
            network, this data is distributed in a round-robin fashion grouping
            by number of cores. Roughly speaking, we can keep this data in memory and
            then we do not have to (de-)serialize the data every time we would like
            to execute a target function with a big dataset. For example, when your
            target function has a big dataset shared across all the target function,
            this argument is very useful.

        Returns
        -------
        status: StatusType
            Status of the trial.
        cost: float | list[float]
            Resulting cost(s) of the trial.
        runtime: float
            The time the target function took to run.
        additional_info: dict
            All further additional trial information.
        """
        # The kwargs are passed to the target function.
        kwargs: dict[str, Any] = {}
        kwargs.update(dask_data_to_scatter)

        if "seed" in self._required_arguments:
            kwargs["seed"] = seed

        if "instance" in self._required_arguments:
            kwargs["instance"] = instance

        if "budget" in self._required_arguments:
            kwargs["budget"] = budget

        # Presetting
        cost: float | list[float] = self._crash_cost
        runtime = 0.0
        additional_info = {}
        status = StatusType.CRASHED

        target_function: Callable
        target_function = self._target_function

        # We don't want the user to change the configuration
        config_copy = copy.deepcopy(config)

        # Call target function
        try:
            # Parse the data yielded by solver/runsolver
            result = self(config_copy, target_function, kwargs)
            status = SMAC3.convert_status(result["status"])
            del result["status"]
            runtime = result["wall_time"]
            del result["wall_time"]
            cpu_time = result["cpu_time"]
            del result["cpu_time"]
            del result["memory"]
            result = {
                key: value for key, value in result.items() if key in self._objectives
            }
        except Exception as e:
            cost = np.asarray(cost).squeeze().tolist()
            additional_info = {
                "traceback": traceback.format_exc(),
                "error": repr(e),
            }
            status = StatusType.CRASHED

        if status == StatusType.CRASHED:
            return status, cost, runtime, cpu_time, additional_info

        # Do some sanity checking (for multi objective)
        error = (
            f"Returned costs {result} does not match "
            f"the number of objectives {self._objectives}."
        )

        # If dict convert to array and make sure the order is correct
        if isinstance(result, dict):
            if len(result) != len(self._objectives):
                raise RuntimeError(error)

            ordered_cost: list[float] = []
            for name in self._objectives:
                if name not in result:
                    raise RuntimeError(
                        f"Objective {name} was not found in the returned costs."
                    )

                ordered_cost.append(result[name])

            result = ordered_cost

        if isinstance(result, list):
            if len(result) != len(self._objectives):
                raise RuntimeError(error)

        if isinstance(result, float):
            if isinstance(self._objectives, list) and len(self._objectives) != 1:
                raise RuntimeError(error)

        cost = result

        if cost is None:
            status = StatusType.CRASHED
            cost = self.crash_cost

        # We want to get either a float or a list of floats.
        cost = np.asarray(cost).squeeze().tolist()

        return status, cost, runtime, cpu_time, additional_info

    def __call__(
        self: SparkleTargetFunctionRunner,
        config: Configuration,
        algorithm: Callable,
        algorithm_kwargs: dict[str, Any],
    ) -> (
        float
        | list[float]
        | dict[str, float]
        | tuple[float, dict]
        | tuple[list[float], dict]
        | tuple[dict[str, float], dict]
    ):
        """Calls the algorithm, which is processed in the ``run`` method."""
        return algorithm(config, **algorithm_kwargs)


if __name__ == "__main__":
    # Incoming call from Sparkle:
    args = sys.argv[1:]
    scenario_file = Path(args[0])
    config_id = args[1]
    seed = int(args[2])
    output_path = Path(args[3])
    scenario = SMAC3Scenario.from_file(scenario_file, run_index=seed)
    global solver, objectives, cutoff, log_dir
    solver = scenario.solver
    cutoff = scenario.solver_cutoff_time
    objectives = scenario.sparkle_objectives
    log_dir = scenario.log_dir
    kwargs = {}
    if scenario.max_ratio is not None:  # Override the default initial design
        kwargs["initial_design"] = scenario.smac_facade.get_initial_design(
            scenario=scenario.smac3_scenario, max_ratio=scenario.max_ratio
        )
    # Facade Configurable
    smac_facade = scenario.smac_facade(
        scenario.smac3_scenario, smac3_solver_call, **kwargs
    )
    # Override the target function runner to control resource management
    smac_facade._runner = SparkleTargetFunctionRunner(
        scenario.smac3_scenario,
        required_arguments=smac_facade._get_signature_arguments(),
    )
    # Refresh the optimiser with new target class
    smac_facade._optimizer = smac_facade._get_optimizer()

    incumbent = smac_facade.optimize()
    # TODO: Fix taking first objective, how do we determine 'best configuration' from
    # a multi objective run?
    SMAC3.organise_output(
        scenario.smac3_scenario.output_directory / "runhistory.json",
        output_target=output_path,
        scenario=scenario,
        configuration_id=config_id,
    )



#!/usr/bin/env python3
"""Sparkle command to execute ablation analysis."""

import argparse
import sys

from runrunner.base import Runner

from sparkle.configurator import AblationScenario
from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import Settings
from sparkle.solver import Solver
from sparkle.structures import PerformanceDataFrame
from sparkle.instance import Instance_Set, InstanceSet
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_object_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Runs parameter importance between the default and configured "
        "parameters with ablation. This command requires a finished "
        "configuration for the solver instance pair.",
        epilog="Note that if no test instance set is given, the validation is performed"
        " on the training set.",
    )
    parser.add_argument("--solver", required=False, type=str, help="path to solver")
    parser.add_argument(
        *ac.InstanceSetTrainOptionalArgument.names,
        **ac.InstanceSetTrainOptionalArgument.kwargs,
    )
    parser.add_argument(
        *ac.InstanceSetTestAblationArgument.names,
        **ac.InstanceSetTestAblationArgument.kwargs,
    )
    # Settings arguments
    parser.add_argument(*ac.SettingsFileArgument.names, **ac.SettingsFileArgument.kwargs)
    parser.add_argument(
        *Settings.OPTION_ablation_racing.args, **Settings.OPTION_ablation_racing.kwargs
    )
    parser.add_argument(*Settings.OPTION_run_on.args, **Settings.OPTION_run_on.kwargs)
    parser.set_defaults(ablation_settings_help=False)
    return parser


def main(argv: list[str]) -> None:
    """Main function to run ablation analysis."""
    sl.log_command(sys.argv, gv.settings().random_state)
    check_for_initialise()

    if not AblationScenario.check_requirements(verbose=True):
        print("Ablation Analysis is not available.")
        if not AblationScenario.ablation_executable.exists():
            print("Would you like to download it? (Y/n)")
            if input().lower().strip() == "y":
                AblationScenario.download_requirements()
            else:
                sys.exit()
        else:
            print("Check that Java is available on your system.")
            sys.exit(-1)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    settings = gv.settings(args)

    # Compare current settings to latest.ini
    prev_settings = Settings(Settings.DEFAULT_previous_settings_path)
    Settings.check_settings_changes(settings, prev_settings)

    solver = resolve_object_name(
        args.solver, gv.solver_nickname_mapping, settings.DEFAULT_solver_dir, Solver
    )
    if solver is None:
        print(f"Could not resolve Solver path/name {args.solver}!")
        print([p for p in settings.DEFAULT_solver_dir.iterdir()])
        sys.exit(-1)

    instance_set_train: InstanceSet = resolve_object_name(
        args.instance_set_train,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        settings.DEFAULT_instance_dir,
        Instance_Set,
    )
    instance_set_test = resolve_object_name(
        args.instance_set_test,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        settings.DEFAULT_instance_dir,
        Instance_Set,
    )

    configurator = settings.configurator
    output_path = settings.get_configurator_output_path(configurator)

    config_scenario = configurator.scenario_class().find_scenario(
        output_path, solver, instance_set_train
    )

    performance_data = PerformanceDataFrame(settings.DEFAULT_performance_data_path)
    if config_scenario is None:
        print(
            "No configuration scenario found for combination:\n"
            f"{configurator.name} {solver.name} {instance_set_train.name}"
        )
        sys.exit(-1)
    best_configuration_key, _ = performance_data.best_configuration(
        str(solver.directory),
        config_scenario.sparkle_objective,
        instances=instance_set_train.instance_names,
    )
    best_configuration = performance_data.get_full_configuration(
        str(solver.directory), best_configuration_key
    )
    if instance_set_test is None:
        instance_set_test = instance_set_train

    if not config_scenario.results_directory.is_dir():
        print(
            "Error: No configuration results found for the given solver and training"
            " instance set. Ablation needs to have a target configuration. "
            "Please finish configuration first."
        )
        sys.exit(-1)
    else:
        print("Configuration exists!")

    ablation_scenario = AblationScenario(
        config_scenario,
        instance_set_test,
        cutoff_length=settings.smac2_target_cutoff_length,  # NOTE: SMAC2
        concurrent_clis=settings.slurm_jobs_in_parallel,
        best_configuration=best_configuration,
        ablation_racing=settings.ablation_racing_flag,
    )

    # Create scenario
    ablation_scenario.create_scenario(override_dirs=True)

    print("Submiting ablation run...")
    run_on = settings.run_on
    runs = ablation_scenario.submit_ablation(
        log_dir=sl.caller_log_dir,
        sbatch_options=settings.sbatch_settings,
        run_on=run_on,
    )

    if run_on == Runner.LOCAL:
        print("Ablation analysis finished!")
    else:
        job_id_str = ",".join([run.run_id for run in runs])
        print(f"Ablation analysis running through Slurm with job id(s): {job_id_str}")
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Init file for Sparkle."""

from sparkle.__about__ import __version__

__version__ = __version__

from sparkle.configurator import (
    SMAC2,
    SMAC2Scenario,
    SMAC3,
    SMAC3Scenario,
    ParamILS,
    ParamILSScenario,
    IRACE,
    IRACEScenario,
)

from sparkle.instance import (
    FileInstanceSet,
    MultiFileInstanceSet,
    IterableFileInstanceSet,
)

from sparkle.selector import (
    Selector,
    SelectionScenario,
    Extractor,
)

from sparkle.solver import (
    Solver,
)

from sparkle.structures import (
    PerformanceDataFrame,
    FeatureDataFrame,
)

# No import from .tools, not important enough

__all__ = [
    "SMAC2",
    "SMAC2Scenario",
    "SMAC3",
    "SMAC3Scenario",
    "ParamILS",
    "ParamILSScenario",
    "IRACE",
    "IRACEScenario",
    "FileInstanceSet",
    "MultiFileInstanceSet",
    "IterableFileInstanceSet",
    "Selector",
    "SelectionScenario",
    "Extractor",
    "Solver",
    "PerformanceDataFrame",
    "FeatureDataFrame",
]



#!/usr/bin/env python3
"""Sparkle command to run solvers to get their performance data."""

from __future__ import annotations
import random
import sys
import argparse
from pathlib import Path

from runrunner.base import Runner, Run

from sparkle.solver import Solver
from sparkle.instance import Instance_Set
from sparkle.structures import PerformanceDataFrame
from sparkle.types import SparkleObjective, resolve_objective
from sparkle.instance import InstanceSet
from sparkle.platform.settings_objects import Settings
from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.nicknames import resolve_object_name, resolve_instance_name
from sparkle.CLI.initialise import check_for_initialise


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Run solvers on instances to get their performance data."
    )
    parser.add_argument(*ac.SolversArgument.names, **ac.SolversArgument.kwargs)
    parser.add_argument(
        *ac.InstanceSetPathsArgument.names, **ac.InstanceSetPathsArgument.kwargs
    )

    # Mutually exclusive: specific configuration or best configuration
    configuration_group = parser.add_mutually_exclusive_group()
    configuration_group.add_argument(
        *ac.ConfigurationArgument.names, **ac.ConfigurationArgument.kwargs
    )
    configuration_group.add_argument(
        *ac.BestConfigurationArgument.names, **ac.BestConfigurationArgument.kwargs
    )
    configuration_group.add_argument(
        *ac.AllConfigurationArgument.names, **ac.AllConfigurationArgument.kwargs
    )
    parser.add_argument(*ac.ObjectiveArgument.names, **ac.ObjectiveArgument.kwargs)
    parser.add_argument(
        *ac.PerformanceDataJobsArgument.names, **ac.PerformanceDataJobsArgument.kwargs
    )
    # This one is only relevant if the argument above is given
    parser.add_argument(
        *ac.RecomputeRunSolversArgument.names, **ac.RecomputeRunSolversArgument.kwargs
    )
    # Settings arguments
    parser.add_argument(*ac.SettingsFileArgument.names, **ac.SettingsFileArgument.kwargs)
    parser.add_argument(
        *Settings.OPTION_solver_cutoff_time.args,
        **Settings.OPTION_solver_cutoff_time.kwargs,
    )
    parser.add_argument(*Settings.OPTION_run_on.args, **Settings.OPTION_run_on.kwargs)
    return parser


def run_solvers(
    solvers: list[Solver],
    instances: list[str] | list[InstanceSet],
    objectives: list[SparkleObjective],
    seed: int,
    cutoff_time: int,
    configurations: list[list[dict[str, str]]],
    sbatch_options: list[str] = None,
    slurm_prepend: str | list[str] | Path = None,
    log_dir: Path = None,
    run_on: Runner = Runner.SLURM,
) -> list[Run]:
    """Run the solvers.

    Parameters
    ----------
    solvers: list[solvers]
        The solvers to run
    instances: list[str] | list[InstanceSet]
        The instances to run the solvers on
    objectives: list[SparkleObjective]
        The objective values to retrieve from the solvers
    seed: int
        The seed to use
    cutoff_time: int
        The cut off time for the solvers
    configurations: list[list[str]]
        The configurations to use for each solver
    sbatch_options: list[str]
        The sbatch options to use for the solvers
    slurm_prepend: str | list[str] | Path
        The script to prepend to a slurm script
    log_dir: Path
        The directory to use for the logs
    run_on: Runner
        Where to execute the solvers.

    Returns
    -------
    run: runrunner.LocalRun or runrunner.SlurmRun
    """
    runs = []
    # Run the solvers
    for solver, solver_confs in zip(solvers, configurations):
        for conf_index, conf in enumerate(solver_confs):
            if "configuration_id" in conf.keys():
                conf_name = conf["configuration_id"]
            else:
                conf_name = conf_index
            run = solver.run(
                instances=instances,
                objectives=objectives,
                seed=seed,
                configuration=conf,
                cutoff_time=cutoff_time,
                run_on=run_on,
                sbatch_options=sbatch_options,
                slurm_prepend=slurm_prepend,
                log_dir=log_dir,
            )
            if run_on == Runner.LOCAL:
                if isinstance(run, dict):
                    run = [run]
                # TODO: Refactor resolving objective keys
                status_key = [key for key in run[0] if key.lower().startswith("status")][
                    0
                ]
                time_key = [key for key in run[0] if key.lower().startswith("cpu_time")][
                    0
                ]
                for i, solver_output in enumerate(run):
                    print(
                        f"Execution of {solver.name} ({conf_name}) on instance "
                        f"{instances[i]} completed with status "
                        f"{solver_output[status_key]} in {solver_output[time_key]} "
                        f"seconds."
                    )
                print("Running configured solver done!")
            else:
                runs.append(run)
    return runs


def run_solvers_performance_data(
    performance_data: PerformanceDataFrame,
    cutoff_time: int,
    rerun: bool = False,
    solvers: list[Solver] = None,
    instances: list[str] = None,
    sbatch_options: list[str] = None,
    slurm_prepend: str | list[str] | Path = None,
    run_on: Runner = Runner.SLURM,
) -> list[Run]:
    """Run the solvers for the performance data.

    Parameters
    ----------
    performance_data: PerformanceDataFrame
        The performance data
    cutoff_time: int
        The cut off time for the solvers
    rerun: bool
        Run only solvers for which no data is available yet (False) or (re)run all
        solvers to get (new) performance data for them (True)
    solvers: list[solvers]
        The solvers to run. If None, run all found solvers.
    instances: list[str]
        The instances to run the solvers on. If None, run all found instances.
    sbatch_options: list[str]
        The sbatch options to use
    slurm_prepend: str | list[str] | Path
        The script to prepend to a slurm script
    run_on: Runner
        Where to execute the solvers. For available values see runrunner.base.Runner
        enum. Default: "Runner.SLURM".

    Returns
    -------
    run: runrunner.LocalRun or runrunner.SlurmRun
        If the run is local return a QueuedRun object with the information concerning
        the run.
    """
    # List of jobs to do
    jobs = performance_data.get_job_list(rerun=rerun)

    # Edit jobs to incorporate file paths
    jobs_with_paths = []
    for solver, config, instance, run in jobs:
        instance_path = resolve_instance_name(
            instance, gv.settings().DEFAULT_instance_dir
        )
        jobs_with_paths.append((solver, config, instance_path, run))
    jobs = jobs_with_paths

    print(f"Total number of jobs to run: {len(jobs)}")
    # If there are no jobs, stop
    if len(jobs) == 0:
        return None

    if run_on == Runner.LOCAL:
        print("Running the solvers locally")
    elif run_on == Runner.SLURM:
        print("Running the solvers through Slurm")

    solvers = performance_data.solvers if solvers is None else solvers
    if solvers is None:
        solver_keys = performance_data.solvers
        solvers = [Solver(Path(s)) for s in solver_keys]
    else:  # Filter the Solvers
        solver_keys = [str(s.directory) for s in solvers]
        jobs = [j for j in jobs if j[0] in solver_keys]
    # Filter the instances
    if instances is not None:
        jobs = [j for j in jobs if j[2] in instances]
    # Sort the jobs per solver
    solver_jobs = {p_solver: {} for p_solver, _, _, _ in jobs}
    for p_solver, p_config, p_instance, p_run in jobs:
        if p_config not in solver_jobs[p_solver]:
            solver_jobs[p_solver][p_config] = {}
        if p_instance not in solver_jobs[p_solver][p_config]:
            solver_jobs[p_solver][p_config][p_instance] = [p_run]
        else:
            solver_jobs[p_solver][p_config][p_instance].append(p_run)
    runrunner_runs = []
    if run_on == Runner.LOCAL:
        print(f"Cutoff time for each solver run: {cutoff_time} seconds")
    for solver, solver_key in zip(solvers, solver_keys):
        for solver_config in solver_jobs[solver_key].keys():
            solver_instances = solver_jobs[solver_key][solver_config].keys()
            run_ids = [
                solver_jobs[solver_key][solver_config][instance]
                for instance in solver_instances
            ]
            if solver_instances == []:
                print(f"Warning: No jobs for instances found for solver {solver_key}")
                continue
            run = solver.run_performance_dataframe(
                solver_instances,
                performance_data,
                solver_config,
                run_ids=run_ids,
                cutoff_time=cutoff_time,
                sbatch_options=sbatch_options,
                slurm_prepend=slurm_prepend,
                log_dir=sl.caller_log_dir,
                base_dir=sl.caller_log_dir,
                run_on=run_on,
            )
            runrunner_runs.append(run)
            if run_on == Runner.LOCAL:
                # Do some printing?
                pass
    if run_on == Runner.SLURM:
        num_jobs = sum(len(r.jobs) for r in runrunner_runs)
        print(f"Total number of jobs submitted: {num_jobs}")

    return runrunner_runs


def main(argv: list[str]) -> None:
    """Main function of the run solvers command."""
    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    settings = gv.settings(args)

    # Log command call
    sl.log_command(sys.argv, seed=settings.random_state)
    check_for_initialise()

    if args.best_configuration:
        if not args.objective:
            objective = settings.objectives[0]
            print(
                "WARNING: Best configuration requested, but no objective specified. "
                f"Defaulting to first objective: {objective}"
            )
        else:
            objective = resolve_objective(args.objective)

    # Compare current settings to latest.ini
    prev_settings = Settings(Settings.DEFAULT_previous_settings_path)
    Settings.check_settings_changes(settings, prev_settings)

    if args.solvers:
        solvers = [
            resolve_object_name(
                solver_path,
                gv.file_storage_data_mapping[gv.solver_nickname_list_path],
                settings.DEFAULT_solver_dir,
                Solver,
            )
            for solver_path in args.solvers
        ]
    else:
        solvers = [
            Solver(p) for p in settings.DEFAULT_solver_dir.iterdir() if p.is_dir()
        ]

    if args.instance_path:
        instances = [
            resolve_object_name(
                instance_path,
                gv.file_storage_data_mapping[gv.instances_nickname_path],
                settings.DEFAULT_instance_dir,
                Instance_Set,
            )
            for instance_path in args.instance_path
        ]
        # Unpack the sets into instance strings
        instances = [str(path) for set in instances for path in set.instance_paths]
    else:
        instances = None  # TODO: Fix? Or its good like this

    sbatch_options = settings.sbatch_settings
    slurm_prepend = settings.slurm_job_prepend
    # Write settings to file before starting, since they are used in callback scripts
    settings.write_used_settings()
    run_on = settings.run_on
    cutoff_time = settings.solver_cutoff_time
    # Open the performance data csv file
    performance_dataframe = PerformanceDataFrame(settings.DEFAULT_performance_data_path)

    print("Start running solvers ...")
    if args.performance_data_jobs:
        runs = run_solvers_performance_data(
            performance_data=performance_dataframe,
            solvers=solvers,
            instances=instances,
            cutoff_time=cutoff_time,
            rerun=args.recompute,
            sbatch_options=sbatch_options,
            slurm_prepend=slurm_prepend,
            run_on=run_on,
        )
    else:
        if args.best_configuration:
            train_instances = None
            if isinstance(args.best_configuration, list):
                train_instances = [
                    resolve_object_name(
                        instance_path,
                        gv.file_storage_data_mapping[gv.instances_nickname_path],
                        settings.DEFAULT_instance_dir,
                        Instance_Set,
                    )
                    for instance_path in args.best_configuration
                ]
                # Unpack the sets into instance strings
                instances = [
                    str(path) for set in train_instances for path in set.instance_paths
                ]
            # Determine best configuration
            configurations = [
                [
                    performance_dataframe.best_configuration(
                        str(solver.directory), objective, train_instances
                    )[0]
                ]
                for solver in solvers
            ]
        elif args.configuration:
            # Sort the configurations to the solvers
            # TODO: Add a better check that the id could only match this solver
            configurations = []
            for solver in solvers:
                configurations.append([])
                for c in args.configuration:
                    if c not in performance_dataframe.configuration_ids:
                        raise ValueError(f"Configuration id {c} not found.")
                    if c in performance_dataframe.get_configurations(
                        str(solver.directory)
                    ):
                        configurations[-1].append(c)
        elif args.all_configurations:  # All known configurations
            configurations = [
                performance_dataframe.get_configurations(str(solver.directory))
                for solver in solvers
            ]
        else:  # Only default configurations
            configurations = [
                [PerformanceDataFrame.default_configuration] for _ in solvers
            ]
        # Look up and replace with the actual configurations
        for solver_index, configs in enumerate(configurations):
            for config_index, config in enumerate(configs):
                configurations[solver_index][config_index] = (
                    performance_dataframe.get_full_configuration(
                        str(solvers[solver_index].directory), config
                    )
                )
        if instances is None:
            instances = []
            for instance_dir in settings.DEFAULT_instance_dir.iterdir():
                if instance_dir.is_dir():
                    instances.append(Instance_Set(instance_dir))

        # TODO Objective arg not used in Multi-file-instances case?
        runs = run_solvers(
            solvers=solvers,
            configurations=configurations,
            instances=instances,
            objectives=settings.objectives,
            seed=random.randint(0, 2**32 - 1),
            cutoff_time=cutoff_time,
            sbatch_options=sbatch_options,
            slurm_prepend=slurm_prepend,
            log_dir=sl.caller_log_dir,
            run_on=run_on,
        )

    # If there are no jobs return
    if runs is None or all(run is None for run in runs):
        print("Running solvers done!")
    elif run_on == Runner.SLURM:
        print(
            "Running solvers through Slurm with job id(s): "
            f"{','.join(r.run_id for r in runs if r is not None)}"
        )
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Sparkle command to generate a report for an executed experiment."""

import sys
import shutil
import argparse
from pathlib import Path
import time
import json
import pandas as pd

from pylatex import NoEscape, NewPage
import pylatex as pl
from sparkle import __version__ as __sparkle_version__

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import resolve_object_name
from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import argparse_custom as ac

from sparkle.solver import Solver
from sparkle.instance import Instance_Set
from sparkle.selector import Extractor
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.configurator.configurator import ConfigurationScenario
from sparkle.selector.selector import SelectionScenario
from sparkle.types import SolverStatus
from sparkle.platform import Settings

from sparkle.platform import latex
from sparkle.platform.output.configuration_output import ConfigurationOutput
from sparkle.platform.output.selection_output import SelectionOutput


MAX_DEC = 4  # Maximum decimals used for each reported value
MAX_COLS_PER_TABLE = 2  # number of value columns extra to number of key columns
WIDE_TABLE_THRESHOLD = 4  # columns above which we switch to landscape
NUM_KEYS_PDF = 3
NUM_KEYS_FDF = 3
MAX_CELL_LEN = 17


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Generates a report for all known selection, configuration and "
        "parallel portfolio scenarios will be generated.",
        epilog="If you wish to filter specific solvers, instance sets, ... have a look "
        "at the command line arguments.",
    )
    # Add argument for filtering solvers
    parser.add_argument(
        *ac.SolversReportArgument.names, **ac.SolversReportArgument.kwargs
    )
    # Add argument for filtering instance sets
    parser.add_argument(
        *ac.InstanceSetsReportArgument.names, **ac.InstanceSetsReportArgument.kwargs
    )

    # Add argument for filtering appendix
    parser.add_argument(
        *Settings.OPTION_appendices.args, **Settings.OPTION_appendices.kwargs
    )

    # Add argument for filtering configurators?
    # Add argument for filtering selectors?
    # Add argument for filtering ??? scenario ids? configuration ids?
    parser.add_argument(*ac.GenerateJSONArgument.names, **ac.GenerateJSONArgument.kwargs)
    return parser


def generate_configuration_section(
    report: pl.Document,
    scenario: ConfigurationScenario,
    scenario_output: ConfigurationOutput,
) -> None:
    """Generate a section for a configuration scenario."""
    report_dir = Path(report.default_filepath).parent
    time_stamp = time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
    plot_dir = (
        report_dir
        / f"{scenario.configurator.__name__}_{scenario.name}_plots_{time_stamp}"
    )
    plot_dir.mkdir(exist_ok=True)

    # 1. Write section intro
    report.append(
        pl.Section(
            f"{scenario.configurator.__name__} Configuration: "
            f"{scenario.solver.name} on {scenario.instance_set.name}"
        )
    )
    report.append("In this scenario, ")
    report.append(
        pl.UnsafeCommand(
            f"textbf{{{scenario.configurator.__name__}}} "
            f"({scenario.configurator.full_name})~\\cite"
            f"{{{scenario.configurator.__name__}}} with version "
            f"{scenario.configurator.version} was used for configuration. "
        )
    )
    report.append(
        f"The Solver {scenario.solver} was optimised on training set "
        f"{scenario.instance_set}. The scenario was run {scenario.number_of_runs} "
        f"times independently with different seeds, yielding {scenario.number_of_runs} "
        f"configurations. The cutoff time for the solver was set to "
        f"{scenario.solver_cutoff_time} seconds. The optimised objective is "
        f"{scenario.sparkle_objectives[0]}. Each Configuration was evaluated on the "
        "training set to determine the best configuration, e.g. the best "
        f"{scenario.sparkle_objectives[0]} value on the training set."
    )

    # 2. Report all the configurator settings in table format
    report.append(pl.Subsection("Configurator Settings"))
    report.append(
        f"The following settings were used for {scenario.configurator.__name__}:\n"
    )
    tabular = pl.Tabular("l|r")
    tabular.add_row("Setting", "Value")
    tabular.add_hline()
    for setting, value in scenario.serialise().items():
        # Keep only the last path segment for paths
        # Otherwise tables get too wide and we can't see other values
        t = str(value).strip().replace("\\", "/")
        parts = [p for p in t.split("/") if p]
        if parts[-1]:
            tabular.add_row([setting, parts[-1]])
        else:
            tabular.add_row([setting, "None"])
    table_conf_settings = pl.Table(position="h")
    table_conf_settings.append(pl.UnsafeCommand("centering"))
    table_conf_settings.append(tabular)
    table_conf_settings.add_caption("Configurator Settings")
    report.append(table_conf_settings)

    # 3. Report details on instance and solver used
    report.append(pl.Subsection("Solver & Instance Set(s) Details"))
    cs = scenario_output.solver.get_configuration_space()
    report.append(
        f"The solver {scenario_output.solver} was configured using "
        f"{len(cs.values())} configurable (hyper)parameters. "
        f"The configuration space has {len(cs.conditions)} conditions. "
    )
    report.append("The following instance sets were used for the scenario:")
    with report.create(pl.Itemize()) as instance_set_latex_list:
        for instance_set in [
            scenario_output.instance_set_train
        ] + scenario_output.test_instance_sets:
            training_set_name = instance_set.name.replace("_", " ")  # Latex fix
            instance_set_latex_list.add_item(
                pl.UnsafeCommand(
                    f"textbf{{{training_set_name}}} ({instance_set.size} instances)"
                )
            )

    # Function to generate a results summary of default vs best on an instance set
    def instance_set_summary(instance_set_name: str) -> None:
        """Generate a results summary of default vs best on an instance set."""
        instance_set_results = scenario_output.instance_set_results[instance_set_name]
        report.append(
            f"The {scenario.sparkle_objectives[0]} value of the Default "
            f"Configuration on {instance_set_name} was "
        )
        report.append(
            pl.UnsafeCommand(
                f"textbf{{{round(instance_set_results.default_performance, MAX_DEC)}}}.\n"
            )
        )
        report.append(
            f"The {scenario.sparkle_objectives[0]} value of the Best "
            f"Configuration on {instance_set_name} was "
        )
        report.append(
            pl.UnsafeCommand(
                f"textbf{{{round(instance_set_results.best_performance, MAX_DEC)}}}.\n"
            )
        )
        report.append("In ")
        report.append(latex.AutoRef(f"fig:bestvsdefault{instance_set_name}{time_stamp}"))
        report.append(pl.utils.bold(" "))  # Force white space
        report.append("the results are plotted per instance.")
        # Create graph to compare best configuration vs default on the instance set

        df = pd.DataFrame(
            [
                instance_set_results.default_instance_performance,
                instance_set_results.best_instance_performance,
            ],
            index=["Default Configuration", "Best Configuration"],
            dtype=float,
        ).T
        plot = latex.comparison_plot(df, None)
        plot_path = (
            plot_dir / f"{scenario_output.best_configuration_key}_vs_"
            f"Default_{instance_set_name}.pdf"
        )
        plot.write_image(plot_path, width=500, height=500)
        with report.create(pl.Figure(position="h")) as figure:
            figure.add_image(
                str(plot_path.relative_to(report_dir)),
                width=pl.utils.NoEscape(r"0.6\textwidth"),
            )
            figure.add_caption(
                f"Best vs Default Performance on {instance_set_name} "
                f"({scenario.sparkle_objectives[0]})"
            )
            figure.append(
                pl.UnsafeCommand(
                    r"label{"
                    f"fig:bestvsdefault{instance_set_name}{time_stamp}"
                    r"}"
                )
            )
        if scenario.sparkle_objectives[0].time:  # Write status table
            report.append("The following Solver status were found per instance:")
            tabular = pl.Tabular("l|c|c|c")
            tabular.add_row("Status", "Default", "Best", "Overlap")
            tabular.add_hline()
            # Count the statuses
            for status in SolverStatus:
                default_count, best_count, overlap_count = 0, 0, 0
                for instance in instance_set_results.instance_status_default.keys():
                    instance = str(instance)
                    default_hit = (
                        instance_set_results.instance_status_default[instance] == status
                    )
                    best_hit = (
                        instance_set_results.instance_status_best[instance] == status
                    )
                    default_count += default_hit
                    best_count += best_hit
                    overlap_count += default_hit and best_hit
                if default_count or best_count:
                    tabular.add_row(status, default_count, best_count, overlap_count)
            table_status_values = pl.Table(position="h")
            table_status_values.append(pl.UnsafeCommand("centering"))
            table_status_values.append(tabular)
            table_status_values.add_caption(
                "Status count for the best and default configuration."
            )
            report.append(table_status_values)

    # 4. Report the results of the best configuration on the training set vs the default
    report.append(
        pl.Subsection(
            f"Comparison of Default and Best Configuration on Training Set "
            f"{scenario_output.instance_set_train.name}"
        )
    )
    instance_set_summary(scenario_output.instance_set_train.name)

    # 5. Report the actual config values
    report.append(pl.Subsubsection("Best Configuration Values"))
    if (
        scenario_output.best_configuration_key
        == PerformanceDataFrame.default_configuration
    ):
        report.append(
            "The configurator failed to find a better configuration than the "
            "default configuration on the training set in this scenario."
        )
    else:
        report.append(
            "The following parameter values "
            "were found to be the best on the training set:\n"
        )
        tabular = pl.Tabular("l|r")
        tabular.add_row("Parameter", "Value")
        tabular.add_hline()
        for parameter, value in scenario_output.best_configuration.items():
            tabular.add_row([parameter, str(value)])
        table_best_values = pl.Table(position="h")
        table_best_values.append(pl.UnsafeCommand("centering"))
        table_best_values.append(tabular)
        table_best_values.add_caption("Best found configuration values")
        report.append(table_best_values)

    # 6. Report the results of best vs default conf on the test sets

    for test_set in scenario_output.test_instance_sets:
        report.append(
            pl.Subsection(
                f"Comparison of Default and Best Configuration on Test Set "
                f"{test_set.name}"
            )
        )
        instance_set_summary(test_set.name)

    # 7. Report the parameter ablation scenario if present
    if scenario.ablation_scenario:
        report.append(pl.Subsection("Parameter importance via Ablation"))
        report.append("Ablation analysis ")
        report.append(pl.UnsafeCommand(r"cite{FawcettHoos16} "))
        test_set = scenario.ablation_scenario.test_set
        if not scenario.ablation_scenario.test_set:
            test_set = scenario.ablation_scenario.train_set
        report.append(
            f"is performed from the default configuration of {scenario.solver} to the "
            f"best found configuration ({scenario_output.best_configuration_key}) "
            "to see which parameter changes between them contribute most to the improved"
            " performance. The ablation path uses the training set "
            f"{scenario.ablation_scenario.train_set.name} and validation is performed "
            f"on the test set {test_set.name}. The set of parameters that differ in the "
            "two configurations will form the ablation path. Starting from the default "
            "configuration, the path is computed by performing a sequence of rounds. In"
            " a round, each available parameter is flipped in the configuration and is "
            "validated on its performance. The flipped parameter with the best "
            "performance in that round, is added to the configuration and the next round"
            " starts with the remaining parameters. This repeats until all parameters "
            "are flipped, which is the best found configuration. The analysis resulted "
            "in the ablation presented in "
        )
        report.append(latex.AutoRef("tab:ablationtable"))
        report.append(".")

        # Add ablation table
        tabular = pl.Tabular("r|l|r|r|r")
        data = scenario.ablation_scenario.read_ablation_table()
        for index, row in enumerate(data):
            tabular.add_row(*row)
            if index == 0:
                tabular.add_hline()
        table_ablation = pl.Table(position="h")
        table_ablation.append(pl.UnsafeCommand("centering"))
        table_ablation.append(tabular)
        table_ablation.add_caption("Ablation table")
        table_ablation.append(pl.UnsafeCommand(r"label{tab:ablationtable}"))
        report.append(table_ablation)


def generate_selection_section(
    report: pl.Document, scenario: SelectionScenario, scenario_output: SelectionOutput
) -> None:
    """Generate a section for a selection scenario."""
    report_dir = Path(report.default_filepath).parent
    time_stamp = time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
    plot_dir = report_dir / f"{scenario.name.replace(' ', '_')}_plots_{time_stamp}"
    plot_dir.mkdir(exist_ok=True)
    report.append(
        pl.Section(
            f"Selection: {scenario.selector.model_class.__name__} on "
            f"{' '.join([s[0] for s in scenario_output.training_instance_sets])}"
        )
    )
    report.append(
        f"In this scenario, a {scenario.selector.model_class.__name__} "
        f" ({scenario.selector.selector_class.__name__}) was trained on the "
        "performance and feature data using ASF-lib. The following solvers "
        f"were run with a cutoff time of {scenario.solver_cutoff} seconds:"
    )
    with report.create(pl.Itemize()) as solver_latex_list:
        for solver_name in scenario_output.solvers.keys():
            solver_name = solver_name.replace("_", " ")
            solver_latex_list.add_item(
                pl.UnsafeCommand(
                    f"textbf{{{solver_name}}} "
                    f"({len(scenario_output.solvers[solver_name])} configurations)"
                )
            )
    # Report training instance sets
    report.append("The following training instance sets were used:")
    with report.create(pl.Itemize()) as instance_set_latex_list:
        for training_set_name, set_size in scenario_output.training_instance_sets:
            training_set_name = training_set_name.replace("_", " ")  # Latex fix
            instance_set_latex_list.add_item(
                pl.UnsafeCommand(f"textbf{{{training_set_name}}} ({set_size} instances)")
            )
    # Report feature extractors
    report.append(
        "The following feature extractors were used with a extractor cutoff "
        f"time of {scenario.extractor_cutoff} seconds:"
    )
    with report.create(pl.Itemize()) as feature_extractor_latex_list:
        for feature_extractor_name in scenario.feature_extractors:
            extractor = resolve_object_name(
                feature_extractor_name,
                gv.file_storage_data_mapping[gv.extractor_nickname_list_path],
                gv.settings().DEFAULT_extractor_dir,
                class_name=Extractor,
            )
            feature_extractor_name = feature_extractor_name.replace("_", " ")  # Latex
            feature_extractor_latex_list.add_item(
                pl.UnsafeCommand(
                    f"textbf{{{feature_extractor_name}}} "
                    f"({extractor.output_dimension} features)"
                )
            )
    # Report Training results
    report.append(pl.Subsection("Training Results"))
    # 1. Report VBS and selector performance,  create ranking list of the solvers
    # TODO Add ref here to the training sets section?
    report.append(
        f"In this section, the {scenario.objective.name} results for the "
        "portfolio selector on solving the training instance set(s) listed "
        "is reported. "
    )
    report.append(
        f"The {scenario.objective.name} values for the Virtual Best Solver "
        "(VBS), i.e., the perfect portfolio selector is "
    )
    report.append(pl.utils.bold(f"{round(scenario_output.vbs_performance, MAX_DEC)}"))
    report.append(", the actual portfolio selector performance is ")
    report.append(
        pl.utils.bold(f"{round(scenario_output.actual_performance, MAX_DEC)}.\n")
    )

    report.append(
        f"Below, the solvers are ranked based on {scenario.objective.name} performance:"
    )
    with report.create(pl.Enumerate()) as ranking_list:
        for solver_name, conf_id, value in scenario_output.solver_performance_ranking:
            value = round(value, MAX_DEC)
            solver_name = solver_name.replace("_", " ")  # Latex fix
            conf_id = conf_id.replace("_", " ")  # Latex fix
            ranking_list.add_item(
                pl.UnsafeCommand(f"textbf{{{solver_name}}} ({conf_id}): {value}")
            )

    # 2. Marginal contribution ranking list VBS
    report.append(pl.Subsubsection("Marginal Contribution Ranking List"))
    report.append(
        "The following list shows the marginal contribution ranking list for the VBS:"
    )
    with report.create(pl.Enumerate()) as ranking_list:
        for (
            solver_name,
            conf_id,
            contribution,
            performance,
        ) in scenario_output.marginal_contribution_perfect:
            contribution, performance = (
                round(contribution, MAX_DEC),
                round(performance, MAX_DEC),
            )
            solver_name = solver_name.replace("_", " ")  # Latex fix
            conf_id = conf_id.replace("_", " ")  # Latex fix
            ranking_list.add_item(
                pl.UnsafeCommand(
                    f"textbf{{{solver_name}}} ({conf_id}): {contribution} ({performance})"
                )
            )

    # 3. Marginal contribution ranking list actual selector
    report.append(
        "The following list shows the marginal contribution ranking list for "
        "the actual portfolio selector:"
    )
    with report.create(pl.Enumerate()) as ranking_list:
        for (
            solver_name,
            conf_id,
            contribution,
            performance,
        ) in scenario_output.marginal_contribution_actual:
            contribution, performance = (
                round(contribution, MAX_DEC),
                round(performance, MAX_DEC),
            )
            solver_name = solver_name.replace("_", " ")  # Latex fix
            conf_id = conf_id.replace("_", " ")  # Latex fix
            ranking_list.add_item(
                pl.UnsafeCommand(
                    f"textbf{{{solver_name}}} ({conf_id}): {contribution} ({performance})"
                )
            )

    # 4. Create scatter plot analysis
    report.append(pl.Subsubsection("Scatter Plot Analysis"))
    report.append(latex.AutoRef(f"fig:sbsvsselector{time_stamp}"))
    report.append(pl.utils.bold(" "))  # Trick to force a white space
    report.append(
        "shows the empirical comparison between the portfolio "
        "selector and the single best solver (SBS). "
    )
    report.append(latex.AutoRef("fig:vbsvsselector"))
    report.append(pl.utils.bold(" "))  # Trick to force a white space
    report.append(
        "shows the empirical comparison between the actual portfolio selector "
        "and the virtual best solver (VBS)."
    )
    # Create figure on SBS versus the selector
    sbs_name, sbs_config, _ = scenario_output.solver_performance_ranking[0]
    # sbs_plot_name = f"{Path(sbs_name).name} ({sbs_config})"
    sbs_performance = scenario_output.sbs_performance
    selector_performance = scenario_output.actual_performance_data

    # Join the data together

    df = pd.DataFrame(
        [sbs_performance, selector_performance],
        index=[f"{Path(sbs_name).name} ({sbs_config})", "Selector"],
        dtype=float,
    ).T
    plot = latex.comparison_plot(df, "Single Best Solver vs Selector")
    plot_path = (
        plot_dir / f"{Path(sbs_name).name}_{sbs_config}_vs_"
        f"Selector_{scenario.selector.model_class.__name__}.pdf"
    )
    plot.write_image(plot_path, width=500, height=500)
    with report.create(pl.Figure()) as figure:
        figure.add_image(
            str(plot_path.relative_to(report_dir)),
            width=pl.utils.NoEscape(r"0.6\textwidth"),
        )
        figure.add_caption(
            "Empirical comparison between the Single Best Solver and the Selector"
        )
        label = r"label{fig:sbsvsselector" + str(time_stamp) + r"}"
        figure.append(pl.UnsafeCommand(f"{label}"))

    # Comparison between the actual portfolio selector in Sparkle and the VBS.
    vbs_performance = scenario_output.vbs_performance_data.tolist()
    df = pd.DataFrame(
        [vbs_performance, selector_performance],
        index=["Virtual Best Solver", "Selector"],
        dtype=float,
    ).T
    plot = latex.comparison_plot(df, "Virtual Best Solver vs Selector")
    plot_path = (
        plot_dir
        / f"Virtual_Best_Solver_vs_Selector_{scenario.selector.model_class.__name__}.pdf"
    )
    plot.write_image(plot_path, width=500, height=500)
    with report.create(pl.Figure()) as figure:
        figure.add_image(
            str(plot_path.relative_to(report_dir)),
            width=pl.utils.NoEscape(r"0.6\textwidth"),
        )
        figure.add_caption(
            "Empirical comparison between the Virtual Best Solver and the Selector"
        )
        figure.append(pl.UnsafeCommand(r"label{fig:vbsvsselector}"))

    if scenario_output.test_sets:
        report.append(pl.Subsection("Test Results"))
        report.append("The following results are reported on the test set(s):")
        with report.create(pl.Itemize()) as latex_list:
            for test_set_name, test_set_size in scenario_output.test_sets:
                result = round(
                    scenario_output.test_set_performance[test_set_name], MAX_DEC
                )
                latex_list.add_item(
                    pl.UnsafeCommand(
                        f"textbf{{{test_set_name}}} ({test_set_size} instances): {result}"
                    )
                )


def generate_parallel_portfolio_section(
    report: pl.Document, scenario: PerformanceDataFrame
) -> None:
    """Generate a section for a parallel portfolio scenario."""
    report_dir = Path(report.default_filepath).parent
    portfolio_name = scenario.csv_filepath.parent.name
    time_stamp = time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
    plot_dir = report_dir / f"{portfolio_name.replace(' ', '_')}_plots_{time_stamp}"
    plot_dir.mkdir()
    report.append(pl.Section(f"Parallel Portfolio {portfolio_name}"))
    report.append(
        "In this scenario, Sparkle runs the portfolio of Solvers on each instance in "
        "parallel with "
        f"{gv.settings().parallel_portfolio_num_seeds_per_solver} different "
        "seeds. The cutoff time for each solver run is set to "
        f"{gv.settings().solver_cutoff_time} seconds."
    )
    report.append(pl.Subsection("Solvers & Instance Sets"))
    report.append("The following Solvers were used in the portfolio:")
    # 1. Report on the Solvers and Instance Sets used for the portfolio
    with report.create(pl.Itemize()) as solver_latex_list:
        configs = scenario.configurations
        for solver in scenario.solvers:
            solver_name = solver.replace("_", " ")
            solver_latex_list.add_item(
                pl.UnsafeCommand(
                    f"textbf{{{solver_name}}} ({len(configs[solver])} configurations)"
                )
            )
    report.append("The following Instance Sets were used in the portfolio:")
    instance_sets = set(Path(instance).parent.name for instance in scenario.instances)
    instance_set_count = [
        len([i for i in scenario.instances if Path(i).parent.name == s])
        for s in instance_sets
    ]
    with report.create(pl.Itemize()) as instance_set_latex_list:
        for set_name, set_size in zip(instance_sets, instance_set_count):
            set_name = set_name.replace("_", " ")  # Latex fix
            instance_set_latex_list.add_item(
                pl.UnsafeCommand(f"textbf{{{set_name}}} ({set_size} instances)")
            )
    # 2. List which solver was the best on how many instances
    report.append(pl.Subsection("Portfolio Performance"))
    objective = scenario.objectives[0]
    report.append(
        f"The objective for the portfolio is {objective}. The "
        "following performance of the solvers was found over the instances: "
    )
    best_solver_count = {solver: 0 for solver in scenario.solvers}
    for instance in scenario.instances:
        ranking = scenario.get_solver_ranking(objective=objective, instances=[instance])
        best_solver_count[ranking[0][0]] += 1

    with report.create(pl.Itemize()) as latex_list:
        for solver, count in best_solver_count.items():
            solver_name = solver.replace("_", " ")
            latex_list.add_item(
                pl.UnsafeCommand(
                    f"textbf{{{solver_name}}} was the best solver on {count} instance(s)."
                )
            )
        # TODO Report how many instances remained unsolved

    # 3. Create table showing the performance of the portfolio vs and all solvers,
    # by showing the status count and number of times the solver was best
    solver_cancelled_count = {solver: 0 for solver in scenario.solvers}
    solver_timeout_count = {solver: 0 for solver in scenario.solvers}
    status_objective = [
        o for o in scenario.objective_names if o.lower().startswith("status")
    ][0]
    cancelled_status = [
        SolverStatus.UNKNOWN,
        SolverStatus.CRASHED,
        SolverStatus.WRONG,
        SolverStatus.ERROR,
        SolverStatus.KILLED,
    ]
    for solver in scenario.solvers:
        status = scenario.get_value(solver=solver, objective=status_objective)
        for status in scenario.get_value(solver=solver, objective=status_objective):
            status = SolverStatus(status)
            if status in cancelled_status:
                solver_cancelled_count[solver] += 1
            elif status == SolverStatus.TIMEOUT:
                solver_timeout_count[solver] += 1

    report.append(latex.AutoRef("tab:parallelportfoliotable"))
    report.append(pl.utils.bold(" "))
    report.append(" shows the performance of the portfolio on the test set(s).")
    tabular = pl.Tabular("r|rrrr")
    tabular.add_row(["Solver", objective, "# Timeouts", "# Cancelled", "# Best"])
    tabular.add_hline()
    solver_performance = {
        solver: round(performance, MAX_DEC)
        for solver, _, performance in scenario.get_solver_ranking(objective=objective)
    }
    for solver in scenario.solvers:
        tabular.add_row(
            solver,
            solver_performance[solver],
            solver_timeout_count[solver],
            solver_cancelled_count[solver],
            best_solver_count[solver],
        )
    tabular.add_hline()
    portfolio_performance = round(
        scenario.best_performance(objective=objective), MAX_DEC
    )
    tabular.add_row(
        portfolio_name,
        portfolio_performance,
        sum(solver_timeout_count.values()),
        sum(solver_cancelled_count.values()),
        sum(best_solver_count.values()),
    )
    table_portfolio = pl.Table(position="h")
    table_portfolio.append(pl.UnsafeCommand("centering"))
    table_portfolio.append(tabular)
    table_portfolio.add_caption("Parallel Portfolio Performance")
    table_portfolio.append(pl.UnsafeCommand(r"label{tab:parallelportfoliotable}"))
    report.append(table_portfolio)

    # 4. Create scatter plot analysis between the portfolio and the single best solver
    sbs_name = scenario.get_solver_ranking(objective=objective)[0][0]
    sbs_instance_performance = scenario.get_value(
        solver=sbs_name, objective=objective.name
    )
    sbs_name = Path(sbs_name).name
    report.append(latex.AutoRef("fig:portfoliovssbs"))
    report.append(pl.utils.bold(" "))
    report.append(
        " shows the emprical comparison between the portfolio and the single "
        f"best solver (SBS) {sbs_name}."
    )
    portfolio_instance_performance = scenario.best_instance_performance(
        objective=objective.name
    ).tolist()

    df = pd.DataFrame(
        [sbs_instance_performance, portfolio_instance_performance],
        index=[f"SBS ({sbs_name}) Performance", "Portfolio Performance"],
        dtype=float,
    ).T
    plot = latex.comparison_plot(df, None)
    plot_path = plot_dir / f"sbs_{sbs_name}_vs_parallel_portfolio.pdf"
    plot.write_image(plot_path, width=500, height=500)
    with report.create(pl.Figure(position="h")) as figure:
        figure.add_image(
            str(plot_path.relative_to(report_dir)),
            width=pl.utils.NoEscape(r"0.6\textwidth"),
        )
        figure.add_caption(f"Portfolio vs SBS Performance ({objective})")
        figure.append(pl.UnsafeCommand(r"label{fig:portfoliovssbs}"))


def append_dataframe_longtable(
    report: pl.Document,
    df: pd.DataFrame,
    caption: str,
    label: str,
    max_cols: int = MAX_COLS_PER_TABLE,
    wide_threshold: int = WIDE_TABLE_THRESHOLD,
    num_keys: int = NUM_KEYS_PDF,
) -> None:
    """Appends a pandas DataFrame to a PyLaTeX document as one or more LaTeX longtables.

    Args:
        report: The PyLaTeX document to which the table(s) will be appended.
        df: The DataFrame to be rendered as LaTeX longtable(s).
        caption: The caption for the table(s).
        label: The LaTeX label for referencing the table(s).
        max_cols: Maximum number of columns per table chunk.
                  Defaults to MAX_COLS_PER_TABLE.
        wide_threshold: Number of columns above which the table is rotated
                        to landscape. Defaults to WIDE_TABLE_THRESHOLD.
        num_keys: Number of key columns to include in each table chunk.
                  Defaults to NUM_KEYS_PDF.

    Returns:
        None
    """
    import math
    from typing import Union

    def latex_escape_text(s: str) -> str:
        """Escape special LaTeX characters in a string."""
        # escape text, but insert our own LaTeX macro around it
        return (
            s.replace("\\", r"\textbackslash{}")
            .replace("&", r"\&")
            .replace("%", r"\%")
            .replace("$", r"\$")
            .replace("#", r"\#")
            .replace("_", r"\_")
            .replace("{", r"\{")
            .replace("}", r"\}")
            .replace("~", r"\textasciitilde{}")
            .replace("^", r"\textasciicircum{}")
        )

    def last_path_segment(text: str) -> str:
        """Keep only the last non-empty path-like segment. Handles both back and forwardslashes. Removes any leading/trailing slashes."""
        t = str(text).strip().replace("\\", "/")
        parts = [p for p in t.split("/") if p]  # ignore empty segments
        return parts[-1] if parts else ""

    def wrap_fixed_shortstack(cell: str, width: int = MAX_CELL_LEN) -> str:
        """Wrap long text to a fixed width for LaTeX tables."""
        string_cell = last_path_segment(cell)
        if len(string_cell) <= width:
            return latex_escape_text(string_cell)
        chunks = [
            latex_escape_text(string_cell[index : index + width])
            for index in range(0, len(string_cell), width)
        ]
        # left-aligned shortstack: forces line breaks and grows row height
        return r"\shortstack[l]{" + r"\\ ".join(chunks) + "}"

    def wrap_header_labels(
        df: pd.DataFrame, width_per_cell: int = MAX_CELL_LEN
    ) -> pd.DataFrame:
        """Wrap long header labels to a fixed width for LaTeX tables."""
        df_copy = df.copy()
        if isinstance(df_copy.columns, pd.MultiIndex):
            new_cols = []
            for tup in df_copy.columns:
                new_cols.append(
                    tuple(
                        wrap_fixed_shortstack(last_path_segment(index), width_per_cell)
                        if isinstance(index, str)
                        else index
                        for index in tup
                    )
                )
            names = [
                (
                    wrap_fixed_shortstack(last_path_segment(name), width_per_cell)
                    if isinstance(name, str)
                    else name
                )
                for name in (df_copy.columns.names or [])
            ]
            df_copy.columns = pd.MultiIndex.from_tuples(new_cols, names=names)
        else:
            df_copy.columns = [
                wrap_fixed_shortstack(last_path_segment(column), width_per_cell)
                if isinstance(column, str)
                else column
                for column in df_copy.columns
            ]
        return df_copy

    def format_cell(cell: Union[int, float, str]) -> str:
        """Format a cell for printing in a LaTeX table."""
        try:
            float_cell = float(cell)
        except (TypeError, ValueError):
            return wrap_fixed_shortstack(last_path_segment(str(cell)), MAX_CELL_LEN)

        if not math.isfinite(float_cell):
            return "NaN"

        if float_cell.is_integer():
            return str(int(float_cell))
        # round to MAX_DEC, then strip trailing zeros
        s = f"{round(float_cell, MAX_DEC):.{MAX_DEC}f}".rstrip("0").rstrip(".")
        return s

    df_copy = df.copy()

    # Inorder to be able to show the key columns, we need to reset the index
    if not isinstance(df_copy.index, pd.RangeIndex) and df_copy.index.name in (
        None,
        "index",
        "",
    ):
        df_copy = df_copy.reset_index()

    # Remove the Seed column from the performance dataframe since it is not
    # very informative and clutters the table
    if isinstance(df, PerformanceDataFrame):
        mask = df_copy.columns.get_level_values("Meta") == "Seed"
        df_copy = df_copy.loc[:, ~mask]

    # For performance dataframe, we want to show values of objectives with their corresponding instance and run.
    # Since objective, instance and run are indexes in the performance dataframe,
    # they will be part of the index and we need to reset the index to get them
    # as columns.
    # We'll name them as key columns, since they are the key to identify the value of the objective
    # for a given instance and run.
    # (Respectively FeatureGroup, FeatureName, Extractor in feature dataframe)
    keys = df_copy.iloc[:, :num_keys]  # Key columns

    # Split the dataframe into chunks of max_cols per page
    number_column_chunks = max((df_copy.shape[1] - 1) // max_cols + 1, 1)
    for i in range(number_column_chunks):
        report.append(NewPage())
        full_part = None
        start_col = i * max_cols
        end_col = (i + 1) * max_cols

        # Select the value columns for this chunk
        values = df_copy.iloc[
            :,
            start_col + num_keys : end_col + num_keys,
        ]

        # Concatenate the key and value columns
        full_part = pd.concat([keys, values], axis=1)

        # If there are no value columns left, we are done
        if (full_part.shape[1]) <= num_keys:
            break

        full_part_wrapped = wrap_header_labels(full_part, MAX_CELL_LEN)

        # tell pandas how to print numbers
        formatters = {col: format_cell for col in full_part_wrapped.columns}

        tex = full_part_wrapped.to_latex(
            longtable=True,
            index=False,
            escape=False,  # We want to split the long words, not escape them
            caption=caption + (f" (part {i + 1})" if number_column_chunks > 1 else ""),
            label=label + f"-p{i + 1}" if number_column_chunks > 1 else label,
            float_format=None,
            multicolumn=True,
            multicolumn_format="c",
            multirow=False,
            column_format="c" * full_part_wrapped.shape[1],
            formatters=formatters,
        )

        # centre the whole table horizontally
        centred_tex = "\\begin{center}\n" + tex + "\\end{center}\n"

        # rotate if still too wide
        if full_part_wrapped.shape[1] > wide_threshold:
            report.append(NoEscape(r"\begin{landscape}"))
            report.append(NoEscape(centred_tex))
            report.append(NoEscape(r"\end{landscape}"))
        else:
            report.append(NoEscape(centred_tex))


def generate_appendix(
    report: pl.Document,
    performance_data: PerformanceDataFrame,
    feature_data: FeatureDataFrame,
) -> None:
    """Appendix.

    Args:
        report: The LaTeX document object to which the appendix will be added.
        performance_data: The performance data to be included in the appendix.
        feature_data: The feature data to be included in the appendix.

    Returns:
        None
    """
    # preamble
    for pkg in ("longtable", "pdflscape", "caption", "booktabs", "placeins"):
        p = pl.Package(pkg)
        if p not in report.packages:
            report.packages.append(p)

    report.append(pl.NewPage())
    report.append(pl.NoEscape(r"\clearpage"))
    report.append(pl.NoEscape(r"\FloatBarrier"))
    report.append(pl.UnsafeCommand("appendix"))
    report.append(pl.Section("Performance DataFrame"))

    append_dataframe_longtable(
        report,
        performance_data,
        caption="Performance DataFrame",
        label="tab:perf_data",
        max_cols=MAX_COLS_PER_TABLE,
        wide_threshold=WIDE_TABLE_THRESHOLD,
        num_keys=NUM_KEYS_PDF,
    )

    report.append(pl.Section("Feature DataFrame"))
    append_dataframe_longtable(
        report,
        feature_data,
        caption="Feature DataFrame",
        label="tab:feature_data",
        max_cols=MAX_COLS_PER_TABLE,
        wide_threshold=WIDE_TABLE_THRESHOLD,
        num_keys=NUM_KEYS_FDF,
    )

    report.append(pl.NoEscape(r"\FloatBarrier"))


def main(argv: list[str]) -> None:
    """Generate a report for executed experiments in the platform."""
    # Log command call
    sl.log_command(sys.argv, gv.settings().random_state)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)

    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)

    # Fetch all known scenarios
    configuration_scenarios = gv.configuration_scenarios(refresh=True)
    selection_scenarios = gv.selection_scenarios(refresh=True)
    parallel_portfolio_scenarios = gv.parallel_portfolio_scenarios()

    # Filter scenarios based on args
    if args.solvers:
        solvers = [
            resolve_object_name(
                s, gv.solver_nickname_mapping, gv.settings().DEFAULT_solver_dir, Solver
            )
            for s in args.solvers
        ]
        configuration_scenarios = [
            s
            for s in configuration_scenarios
            if s.solver.directory in [s.directory for s in solvers]
        ]
        selection_scenarios = [
            s
            for s in selection_scenarios
            if set(s.solvers).intersection([str(s.directory) for s in solvers])
        ]
        parallel_portfolio_scenarios = [
            s
            for s in parallel_portfolio_scenarios
            if set(s.solvers).intersection([str(s.directory) for s in solvers])
        ]
    if args.instance_sets:
        instance_sets = [
            resolve_object_name(
                s,
                gv.instance_set_nickname_mapping,
                gv.settings().DEFAULT_instance_dir,
                Instance_Set,
            )
            for s in args.instance_sets
        ]
        configuration_scenarios = [
            s
            for s in configuration_scenarios
            if s.instance_set.directory in [s.directory for s in instance_sets]
        ]
        selection_scenarios = [
            s
            for s in selection_scenarios
            if set(s.instance_sets).intersection([str(s.name) for s in instance_sets])
        ]
        parallel_portfolio_scenarios = [
            s
            for s in parallel_portfolio_scenarios
            if set(s.instance_sets).intersection([str(s.name) for s in instance_sets])
        ]

    processed_configuration_scenarios = []
    processed_selection_scenarios = []
    possible_test_sets = [
        Instance_Set(p) for p in gv.settings().DEFAULT_instance_dir.iterdir()
    ]
    for configuration_scenario in configuration_scenarios:
        processed_configuration_scenarios.append(
            (
                ConfigurationOutput(
                    configuration_scenario, performance_data, possible_test_sets
                ),
                configuration_scenario,
            )
        )
    for selection_scenario in selection_scenarios:
        processed_selection_scenarios.append(
            (SelectionOutput(selection_scenario), selection_scenario)
        )

    raw_output = gv.settings().DEFAULT_output_analysis / "JSON"
    if raw_output.exists():  # Clean
        shutil.rmtree(raw_output)
    raw_output.mkdir()

    # Write JSON
    output_json = {}
    for output, configuration_scenario in processed_configuration_scenarios:
        output_json[configuration_scenario.name] = output.serialise()
    for output, selection_scenario in processed_selection_scenarios:
        output_json[selection_scenario.name] = output.serialise()
    # TODO: We do not have an output object for parallel portfolios

    raw_output_json = raw_output / "output.json"
    with raw_output_json.open("w") as f:
        json.dump(output_json, f, indent=4)

    print(f"Machine readable output written to: {raw_output_json}")

    if args.only_json:  # Done
        sys.exit(0)

    # TODO: Group scenarios based on:
    # - Configuration / Selection / Parallel Portfolio
    # - Training Instance Set / Testing Instance Set
    # - Configurators can be merged as long as we can match their budgets clearly
    report_directory = gv.settings().DEFAULT_output_analysis / "report"
    if report_directory.exists():  # Clean it
        shutil.rmtree(report_directory)
    report_directory.mkdir()
    target_path = report_directory / "report"
    report = pl.document.Document(
        default_filepath=str(target_path), document_options=["british"]
    )
    bibpath = gv.settings().bibliography_path
    newbibpath = report_directory / "report.bib"
    shutil.copy(bibpath, newbibpath)
    # BUGFIX for unknown package load in PyLatex
    p = pl.package.Package("lastpage")
    if p in report.packages:
        report.packages.remove(p)
    report.packages.append(
        pl.package.Package(
            "geometry",
            options=[
                "verbose",
                "tmargin=3.5cm",
                "bmargin=3.5cm",
                "lmargin=3cm",
                "rmargin=3cm",
            ],
        )
    )
    # Unsafe command for \emph{Sparkle}
    report.preamble.extend(
        [
            pl.UnsafeCommand("title", r"\emph{Sparkle} Algorithm Portfolio report"),
            pl.UnsafeCommand(
                "author",
                r"Generated by \emph{Sparkle} "
                f"(version: {__sparkle_version__})",
            ),
        ]
    )
    report.append(pl.Command("maketitle"))
    report.append(pl.Section("Introduction"))
    # TODO: A quick overview to the introduction on whats considered in the report
    # regarding Solvers, Instance Sets and Feature Extractors
    report.append(
        pl.UnsafeCommand(
            r"emph{Sparkle}~\cite{Hoos15} is a multi-agent problem-solving platform based on"
            r" Programming by Optimisation (PbO)~\cite{Hoos12}, and would provide a number "
            "of effective algorithm optimisation techniques (such as automated algorithm "
            "configuration, portfolio-based algorithm selection, etc.) to accelerate the "
            "existing solvers."
        )
    )

    for scenario_output, scenario in processed_configuration_scenarios:
        generate_configuration_section(report, scenario, scenario_output)

    for scenario_output, scenario in processed_selection_scenarios:
        generate_selection_section(report, scenario, scenario_output)

    for parallel_dataframe in parallel_portfolio_scenarios:
        generate_parallel_portfolio_section(report, parallel_dataframe)

    # Check if user wants to add appendix and
    settings = gv.settings(args)
    if settings.appendices:
        generate_appendix(report, performance_data, feature_data)

    # Adding bibliography
    report.append(pl.NewPage())  # Ensure it starts on new page
    report.append(pl.Command("bibliographystyle", arguments=["plain"]))
    report.append(pl.Command("bibliography", arguments=[str(newbibpath)]))
    # Generate the report .tex and .pdf
    report.generate_pdf(target_path, clean=False, clean_tex=False, compiler="pdflatex")
    # TODO: This should be done by PyLatex. Generate the bib and regenerate the report
    # Reference for the (terrible) solution: https://tex.stackexchange.com/
    # questions/63852/question-mark-or-bold-citation-key-instead-of-citation-number
    import subprocess

    # Run BibTex silently
    subprocess.run(
        ["bibtex", newbibpath.with_suffix("")],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    report.generate_pdf(target_path, clean=False, clean_tex=False, compiler="pdflatex")
    report.generate_pdf(target_path, clean=False, clean_tex=False, compiler="pdflatex")
    print(f"Report generated at {target_path}.pdf")
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""File to handle a solver and its directories."""

from __future__ import annotations
import sys
from typing import Any
import shlex
import ast
import json
import random
from pathlib import Path

from ConfigSpace import ConfigurationSpace

import runrunner as rrr
from runrunner.local import LocalRun
from runrunner.slurm import Run, SlurmRun
from runrunner.base import Status, Runner

from sparkle.tools.parameters import PCSConverter, PCSConvention
from sparkle.tools import RunSolver
from sparkle.types import SparkleCallable, SolverStatus
from sparkle.solver import verifiers
from sparkle.instance import InstanceSet
from sparkle.structures import PerformanceDataFrame
from sparkle.types import resolve_objective, SparkleObjective, UseTime


class Solver(SparkleCallable):
    """Class to handle a solver and its directories."""

    meta_data = "solver_meta.txt"
    _wrapper_file = "sparkle_solver_wrapper"
    solver_cli = Path(__file__).parent / "solver_cli.py"

    def __init__(
        self: Solver,
        directory: Path,
        runsolver_exec: Path = None,
        deterministic: bool = None,
        verifier: verifiers.SolutionVerifier = None,
    ) -> None:
        """Initialize solver.

        Args:
            directory: Directory of the solver.
            runsolver_exec: Path to the runsolver executable.
                By default, runsolver in directory.
            deterministic: Bool indicating determinism of the algorithm.
                Defaults to False.
            verifier: The solution verifier to use. If None, no verifier is used.
        """
        super().__init__(directory, runsolver_exec)
        self.deterministic = deterministic
        self.verifier = verifier
        self._pcs_file: Path = None
        self._interpreter: str = None
        self._wrapper_extension: str = None

        meta_data_file = self.directory / Solver.meta_data
        if meta_data_file.exists():
            meta_data = ast.literal_eval(meta_data_file.open().read())
            # We only override the deterministic and verifier from file if not set
            if self.deterministic is None:
                if (
                    "deterministic" in meta_data
                    and meta_data["deterministic"] is not None
                ):
                    self.deterministic = meta_data["deterministic"]
            if self.verifier is None and "verifier" in meta_data:
                if isinstance(meta_data["verifier"], tuple):  # File verifier
                    self.verifier = verifiers.mapping[meta_data["verifier"][0]](
                        Path(meta_data["verifier"][1])
                    )
                elif meta_data["verifier"] in verifiers.mapping:
                    self.verifier = verifiers.mapping[meta_data["verifier"]]
        if self.deterministic is None:  # Default to False
            self.deterministic = False

    def __str__(self: Solver) -> str:
        """Return the string representation of the solver."""
        return self.name

    def __repr__(self: Solver) -> str:
        """Return detailed representation of the solver."""
        return (
            f"{self.name}:\n"
            f"\t- Directory: {self.directory}\n"
            f"\t- Deterministic: {self.deterministic}\n"
            f"\t- Verifier: {self.verifier}\n"
            f"\t- PCS File: {self.pcs_file}\n"
            f"\t- Wrapper: {self.wrapper}"
        )

    @property
    def pcs_file(self: Solver) -> Path:
        """Get path of the parameter file."""
        if self._pcs_file is None:
            files = sorted([p for p in self.directory.iterdir() if p.suffix == ".pcs"])
            if len(files) == 0:
                return None
            self._pcs_file = files[0]
        return self._pcs_file

    @property
    def wrapper_extension(self: Solver) -> str:
        """Get the extension of the wrapper file."""
        if self._wrapper_extension is None:
            # Determine which file is the wrapper by sorting alphabetically
            wrapper = sorted(
                [p for p in self.directory.iterdir() if p.stem == Solver._wrapper_file]
            )[0]
            self._wrapper_extension = wrapper.suffix
        return self._wrapper_extension

    @property
    def wrapper(self: Solver) -> str:
        """Get name of the wrapper file."""
        return f"{Solver._wrapper_file}{self.wrapper_extension}"

    @property
    def wrapper_file(self: Solver) -> Path:
        """Get path of the wrapper file."""
        return self.directory / self.wrapper

    def get_pcs_file(self: Solver, port_type: PCSConvention) -> Path:
        """Get path of the parameter file of a specific convention.

        Args:
            port_type: Port type of the parameter file. If None, will return the
                file with the shortest name.

        Returns:
            Path to the parameter file. None if it can not be resolved.
        """
        pcs_files = sorted([p for p in self.directory.iterdir() if p.suffix == ".pcs"])
        if port_type is None:
            return pcs_files[0]
        for file in pcs_files:
            if port_type == PCSConverter.get_convention(file):
                return file
        return None

    def read_pcs_file(self: Solver) -> bool:
        """Checks if the pcs file can be read."""
        # TODO: Should be a .validate method instead
        return PCSConverter.get_convention(self.pcs_file) is not None

    def get_configuration_space(self: Solver) -> ConfigurationSpace:
        """Get the ConfigurationSpace of the PCS file."""
        if not self.pcs_file:
            return None
        return PCSConverter.parse(self.pcs_file)

    def port_pcs(self: Solver, port_type: PCSConvention) -> None:
        """Port the parameter file to the given port type."""
        target_pcs_file = (
            self.pcs_file.parent / f"{self.pcs_file.stem}_{port_type.name}.pcs"
        )
        if target_pcs_file.exists():  # Already exists, possibly user defined
            return
        PCSConverter.export(self.get_configuration_space(), port_type, target_pcs_file)

    def build_cmd(
        self: Solver,
        instance: str | list[str],
        objectives: list[SparkleObjective],
        seed: int,
        cutoff_time: int = None,
        configuration: dict = None,
        log_dir: Path = None,
    ) -> list[str]:
        """Build the solver call on an instance with a configuration.

        Args:
            instance: Path to the instance.
            objectives: List of sparkle objectives.
            seed: Seed of the solver.
            cutoff_time: Cutoff time for the solver.
            configuration: Configuration of the solver.
            log_dir: Directory path for logs.

        Returns:
            List of commands and arguments to execute the solver.
        """
        if configuration is None:
            configuration = {}
        # Ensure configuration contains required entries for each wrapper
        configuration["solver_dir"] = str(self.directory.absolute())
        configuration["instance"] = instance
        configuration["seed"] = seed
        configuration["objectives"] = ",".join([str(obj) for obj in objectives])
        configuration["cutoff_time"] = (
            cutoff_time if cutoff_time is not None else sys.maxsize
        )
        if "configuration_id" in configuration:
            del configuration["configuration_id"]
        # Ensure stringification of dictionary will go correctly for key value pairs
        configuration = {key: str(configuration[key]) for key in configuration}
        solver_cmd = [
            str(self.directory / self.wrapper),
            f"'{json.dumps(configuration)}'",
        ]
        if log_dir is None:
            log_dir = Path()
        if cutoff_time is not None:  # Use RunSolver
            log_path_str = instance[0] if isinstance(instance, list) else instance
            log_name_base = f"{Path(log_path_str).name}_{self.name}"
            return RunSolver.wrap_command(
                self.runsolver_exec,
                solver_cmd,
                cutoff_time,
                log_dir,
                log_name_base=log_name_base,
            )
        return solver_cmd

    def run(
        self: Solver,
        instances: str | list[str] | InstanceSet | list[InstanceSet],
        objectives: list[SparkleObjective],
        seed: int,
        cutoff_time: int = None,
        configuration: dict = None,
        run_on: Runner = Runner.LOCAL,
        sbatch_options: list[str] = None,
        slurm_prepend: str | list[str] | Path = None,
        log_dir: Path = None,
    ) -> SlurmRun | list[dict[str, Any]] | dict[str, Any]:
        """Run the solver on an instance with a certain configuration.

        Args:
            instances: The instance(s) to run the solver on, list in case of multi-file.
                In case of an instance set, will run on all instances in the set.
            objectives: List of sparkle objectives.
            seed: Seed to run the solver with. Fill with abitrary int in case of
                determnistic solver.
            cutoff_time: The cutoff time for the solver, measured through RunSolver.
                If None, will be executed without RunSolver.
            configuration: The solver configuration to use. Can be empty.
            run_on: Whether to run on slurm or locally.
            sbatch_options: The sbatch options to use.
            slurm_prepend: The script to prepend to a slurm script.
            log_dir: The log directory to use.

        Returns:
            Solver output dict possibly with runsolver values.
        """
        cmds = []
        set_label = instances.name if isinstance(instances, InstanceSet) else "instances"
        instances = [instances] if not isinstance(instances, list) else instances
        log_dir = Path() if log_dir is None else log_dir

        for instance in instances:
            paths = (
                instance.instance_paths
                if isinstance(instance, InstanceSet)
                else [instance]
            )
            for instance_path in paths:
                instance_path = (
                    [str(p) for p in instance_path]
                    if isinstance(instance_path, list)
                    else instance_path
                )
                solver_cmd = self.build_cmd(
                    instance_path,
                    objectives=objectives,
                    seed=seed,
                    cutoff_time=cutoff_time,
                    configuration=configuration,
                    log_dir=log_dir,
                )
                cmds.append(" ".join(solver_cmd))

        commandname = f"Run Solver: {self.name} on {set_label}"
        run = rrr.add_to_queue(
            runner=run_on,
            cmd=cmds,
            name=commandname,
            base_dir=log_dir,
            sbatch_options=sbatch_options,
            prepend=slurm_prepend,
        )

        if isinstance(run, LocalRun):
            run.wait()
            if run.status == Status.ERROR:  # Subprocess resulted in error
                print(f"WARNING: Solver {self.name} execution seems to have failed!\n")
                for i, job in enumerate(run.jobs):
                    print(
                        f"[Job {i}] The used command was: {cmds[i]}\n"
                        "The error yielded was:\n"
                        f"\t-stdout: '{job.stdout}'\n"
                        f"\t-stderr: '{job.stderr}'\n"
                    )
                return {
                    "status": SolverStatus.ERROR,
                }

            solver_outputs = []
            for i, job in enumerate(run.jobs):
                solver_cmd = cmds[i].split(" ")
                solver_output = Solver.parse_solver_output(
                    run.jobs[i].stdout,
                    solver_call=solver_cmd,
                    objectives=objectives,
                    verifier=self.verifier,
                )
                solver_outputs.append(solver_output)
            return solver_outputs if len(solver_outputs) > 1 else solver_output
        return run

    def run_performance_dataframe(
        self: Solver,
        instances: str | list[str] | InstanceSet,
        performance_dataframe: PerformanceDataFrame,
        config_ids: str | list[str] = None,
        run_ids: list[int] | list[list[int]] = None,
        cutoff_time: int = None,
        objective: SparkleObjective = None,
        train_set: InstanceSet = None,
        sbatch_options: list[str] = None,
        slurm_prepend: str | list[str] | Path = None,
        dependencies: list[SlurmRun] = None,
        log_dir: Path = None,
        base_dir: Path = None,
        job_name: str = None,
        run_on: Runner = Runner.SLURM,
    ) -> Run:
        """Run the solver from and place the results in the performance dataframe.

        This in practice actually runs Solver.run, but has a little script before/after,
        to read and write to the performance dataframe.

        Args:
            instances: The instance(s) to run the solver on. In case of an instance set,
                or list, will create a job for all instances in the set/list.
            config_ids: The config indices to use in the performance dataframe.
            performance_dataframe: The performance dataframe to use.
            run_ids: List of run ids to use. If list of list, a list of runs is given
                per instance. Otherwise, all runs are used for each instance.
            cutoff_time: The cutoff time for the solver, measured through RunSolver.
            objective: The objective to use, only relevant when determining the best
                configuration.
            train_set: The training set to use. If present, will determine the best
                configuration of the solver using these instances and run with it on
                all instances in the instance argument.
            sbatch_options: List of slurm batch options to use
            slurm_prepend: Slurm script to prepend to the sbatch
            dependencies: List of slurm runs to use as dependencies
            log_dir: Path where to place output files. Defaults to CWD.
            base_dir: Path where to place output files.
            job_name: Name of the job
                If None, will generate a name based on Solver and Instances
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            SlurmRun or Local run of the job.
        """
        instances = [instances] if isinstance(instances, str) else instances
        set_name = "instances"
        if isinstance(instances, InstanceSet):
            set_name = instances.name
            instances = [str(i) for i in instances.instance_paths]
        if not isinstance(config_ids, list):
            config_ids = [config_ids]
        configurations = [
            performance_dataframe.get_full_configuration(str(self.directory), config_id)
            if config_id
            else None
            for config_id in config_ids
        ]
        if run_ids is None:
            run_ids = performance_dataframe.run_ids
        if isinstance(run_ids[0], list):  # Runs per instance
            combinations = []
            for index, instance in enumerate(instances):
                for run_id in run_ids[index]:
                    combinations.extend(
                        [
                            (instance, config_id, config, run_id)
                            for config_id, config in zip(config_ids, configurations)
                        ]
                    )
        else:  # Runs for all instances
            import itertools

            combinations = [
                (instance, config_data[0], config_data[1], run_id)
                for instance, config_data, run_id in itertools.product(
                    instances,
                    zip(config_ids, configurations),
                    performance_dataframe.run_ids,
                )
            ]
        objective_arg = f"--target-objective {objective.name}" if objective else ""
        train_arg = (
            "--best-configuration-instances "
            + " ".join([str(i) for i in train_set.instance_paths])
            if train_set
            else ""
        )
        configuration_args = [
            ""
            if not config_id and not config
            else f"--configuration-id {config_id}"
            if not config
            else f"--configuration '{json.dumps(config)}'"
            for _, config_id, config, _ in combinations
        ]

        # We run all instances/configs/runs combinations
        # For each value we try to resolve from the PDF, to avoid high read loads during executions
        cmds = [
            f"python3 {Solver.solver_cli} "
            f"--solver {self.directory} "
            f"--instance {instance} "
            f"{config_arg} "
            # f"{'--configuration-id ' + config_id if not config else '--configuration"' + str(config) + '\"'} "
            f"--run-index {run_id} "
            f"--objectives {' '.join([obj.name for obj in performance_dataframe.objectives])} "
            f"--performance-dataframe {performance_dataframe.csv_filepath} "
            f"--cutoff-time {cutoff_time} "
            f"--log-dir {log_dir} "
            f"--seed {random.randint(0, 2**32 - 1)} "
            f"{objective_arg} "
            f"{train_arg}"
            for (instance, _, _, run_id), config_arg in zip(
                combinations, configuration_args
            )
        ]
        job_name = f"Run: {self.name} on {set_name}" if job_name is None else job_name
        r = rrr.add_to_queue(
            runner=run_on,
            cmd=cmds,
            name=job_name,
            base_dir=base_dir,
            sbatch_options=sbatch_options,
            prepend=slurm_prepend,
            dependencies=dependencies,
        )
        if run_on == Runner.LOCAL:
            r.wait()
        return r

    @staticmethod
    def config_str_to_dict(config_str: str) -> dict[str, str]:
        """Parse a configuration string to a dictionary."""
        # First we filter the configuration of unwanted characters
        config_str = config_str.strip().replace("-", "")
        # Then we split the string by spaces, but conserve substrings
        config_list = shlex.split(config_str)
        # We return empty for empty input OR uneven input
        if config_str == "" or config_str == r"{}" or len(config_list) & 1:
            return {}
        config_dict = {}
        for index in range(0, len(config_list), 2):
            # As the value will already be a string object, no quotes are allowed in it
            value = config_list[index + 1].strip('"').strip("'")
            config_dict[config_list[index]] = value
        return config_dict

    @staticmethod
    def parse_solver_output(
        solver_output: str,
        solver_call: list[str | Path] = None,
        objectives: list[SparkleObjective] = None,
        verifier: verifiers.SolutionVerifier = None,
    ) -> dict[str, Any]:
        """Parse the output of the solver.

        Args:
            solver_output: The output of the solver run which needs to be parsed
            solver_call: The solver call used to run the solver
            objectives: The objectives to apply to the solver output
            verifier: The verifier to check the solver output

        Returns:
            Dictionary representing the parsed solver output
        """
        used_runsolver = False
        if solver_call is not None and len(solver_call) > 2:
            used_runsolver = True
            parsed_output = RunSolver.get_solver_output(solver_call, solver_output)
        else:
            parsed_output = ast.literal_eval(solver_output)
        # cast status attribute from str to Enum
        parsed_output["status"] = SolverStatus(parsed_output["status"])
        # Apply objectives to parsed output, runtime based objectives added here
        if verifier is not None and used_runsolver:
            # Horrible hack to get the instance from the solver input
            solver_call_str: str = " ".join(solver_call)
            solver_input_str = solver_call_str.split(Solver._wrapper_file, maxsplit=1)[1]
            solver_input_str = solver_input_str.split(" ", maxsplit=1)[1]
            solver_input_str = solver_input_str[
                solver_input_str.index("{") : solver_input_str.index("}") + 1
            ]
            solver_input = ast.literal_eval(solver_input_str)
            target_instance = Path(solver_input["instance"])
            parsed_output["status"] = verifier.verify(
                target_instance, parsed_output, solver_call
            )

        # Create objective map
        objectives = {o.stem: o for o in objectives} if objectives else {}
        removable_keys = ["cutoff_time"]  # Keys to remove

        # apply objectives to parsed output, runtime based objectives added here
        for key, value in parsed_output.items():
            if objectives and key in objectives:
                objective = objectives[key]
                removable_keys.append(key)  # We translate it into the full name
            else:
                objective = resolve_objective(key)
            # If not found in objectives, resolve to which objective the output belongs
            if objective is None:  # Could not parse, skip
                continue
            if objective.use_time == UseTime.NO:
                if objective.post_process is not None:
                    parsed_output[key] = objective.post_process(value)
            else:
                if not used_runsolver:
                    continue
                if objective.use_time == UseTime.CPU_TIME:
                    parsed_output[key] = parsed_output["cpu_time"]
                else:
                    parsed_output[key] = parsed_output["wall_time"]
                if objective.post_process is not None:
                    parsed_output[key] = objective.post_process(
                        parsed_output[key],
                        parsed_output["cutoff_time"],
                        parsed_output["status"],
                    )

        # Replace or remove keys based on the objective names
        for key in removable_keys:
            if key in parsed_output:
                if key in objectives:
                    # Map the result to the objective
                    parsed_output[objectives[key].name] = parsed_output[key]
                    if key != objectives[key].name:  # Only delete actual mappings
                        del parsed_output[key]
                else:
                    del parsed_output[key]
        return parsed_output



"""Package meta file."""

import importlib.metadata

__name__ = "sparkle"
__version__ = importlib.metadata.version(__name__)

description = importlib.metadata.metadata(__name__)["Summary"]
licence = importlib.metadata.metadata(__name__)["License-Expression"]
authors = importlib.metadata.metadata(__name__)["Author"]
contact = importlib.metadata.metadata(__name__)["Maintainer-email"].split(", ")[-1]



"""Sparkle class to organise configuration output."""

from __future__ import annotations
import operator
import json
from pathlib import Path

from sparkle.selector import SelectionScenario
from sparkle.structures import PerformanceDataFrame
from sparkle.platform.output.structures import (
    SelectionPerformance,
    SelectionSolverData,
)


def compute_selector_marginal_contribution(
    selection_scenario: SelectionScenario,
) -> list[tuple[str, float]]:
    """Compute the marginal contributions of solvers in the selector.

    Args:
      performance_data: Performance data object
      feature_data_csv_path: Path to the CSV file with the feature data.
      selection_scenario: The selector scenario for which to compute
        marginal contribution.
      objective: Objective to compute the marginal contribution for.

    Returns:
      A list of 4-tuples where every 4-tuple is of the form
        (solver_name, config_id, marginal contribution, best_performance).
    """
    selector_performance = selection_scenario.objective.instance_aggregator(
        selection_scenario.selector_performance_data.get_value(
            SelectionScenario.__selector_solver_name__,
            instance=selection_scenario.training_instances,
            objective=selection_scenario.objective.name,
        )
    )
    rank_list = []
    compare = operator.lt if selection_scenario.objective.minimise else operator.gt
    # Compute contribution per solver
    for ablation_scenario in selection_scenario.ablation_scenarios:
        # Hacky way of getting the needed data on the ablation
        _, solver_name, config = ablation_scenario.directory.name.split("_", maxsplit=2)
        # Hacky way of reconstructing the solver id in the PDF
        solver = f"Solvers/{solver_name}"
        ablated_selector_performance = ablation_scenario.objective.instance_aggregator(
            ablation_scenario.selector_performance_data.get_value(
                SelectionScenario.__selector_solver_name__,
                instance=ablation_scenario.training_instances,
                objective=ablation_scenario.objective.name,
            )
        )

        # 1. If the performance remains equal, this solver did not contribute
        # 2. If there is a performance decay without this solver, it does contribute
        # 3. If there is a performance improvement, we have a bad portfolio selector
        if ablated_selector_performance == selector_performance:
            marginal_contribution = 0.0
        elif not compare(ablated_selector_performance, selector_performance):
            # The performance decreases, we have a contributing solver
            marginal_contribution = ablated_selector_performance / selector_performance
        else:
            print(
                "****** WARNING DUBIOUS SELECTOR/SOLVER: "
                f"The omission of solver {solver_name} ({config}) yields an "
                "improvement. The selector improves better without this solver. "
                "It may be usefull to construct a portfolio without this solver."
            )
            marginal_contribution = 0.0

        rank_list.append(
            (solver, config, marginal_contribution, ablated_selector_performance)
        )

    rank_list.sort(key=lambda contribution: contribution[2], reverse=True)
    return rank_list


class SelectionOutput:
    """Class that collects selection data and outputs it a JSON format."""

    def __init__(
        self: SelectionOutput,
        selection_scenario: SelectionScenario,
    ) -> None:
        """Initialize SelectionOutput class.

        Args:
            selection_scenario: Path to selection output directory
            performance_data: The performance data used for the selector
        """
        self.training_instances = selection_scenario.training_instances
        training_instance_sets = selection_scenario.training_instance_sets
        self.training_instance_sets = [
            (instance_set, sum(instance_set in s for s in self.training_instances))
            for instance_set in training_instance_sets
        ]
        self.test_instances = selection_scenario.test_instances
        test_sets = selection_scenario.test_instance_sets
        self.test_sets = [
            (instance_set, sum(instance_set in s for s in self.test_instances))
            for instance_set in test_sets
        ]
        self.cutoff_time = selection_scenario.solver_cutoff
        self.objective = selection_scenario.objective

        solver_performance_data = selection_scenario.selector_performance_data.clone()
        solver_performance_data.remove_solver(SelectionScenario.__selector_solver_name__)

        self.solver_performance_ranking = solver_performance_data.get_solver_ranking(
            instances=self.training_instances, objective=self.objective
        )

        self.solver_data = self.get_solver_data(solver_performance_data)
        self.solvers = {}
        for solver_conf in selection_scenario.performance_data.columns:
            solver, conf = solver_conf.split("_", maxsplit=1)
            if solver not in self.solvers:
                self.solvers[solver] = []
            self.solvers[solver].append(conf)

        self.sbs_performance = solver_performance_data.get_value(
            solver=self.solver_performance_ranking[0][0],
            configuration=self.solver_performance_ranking[0][1],
            instance=self.training_instances,
            objective=self.objective.name,
        )

        # Collect marginal contribution data
        self.marginal_contribution_perfect = (
            solver_performance_data.marginal_contribution(
                selection_scenario.objective,
                instances=self.training_instances,
                sort=True,
            )
        )

        self.marginal_contribution_actual = compute_selector_marginal_contribution(
            selection_scenario
        )
        # Collect performance data
        self.vbs_performance_data = solver_performance_data.best_instance_performance(
            instances=self.training_instances, objective=selection_scenario.objective
        )
        self.vbs_performance = selection_scenario.objective.instance_aggregator(
            self.vbs_performance_data
        )

        self.test_set_performance = {} if self.test_sets else None
        for test_set, _ in self.test_sets:
            test_set_instances = [
                instance for instance in self.test_instances if test_set in instance
            ]
            test_perf = selection_scenario.selector_performance_data.best_performance(
                exclude_solvers=[
                    s
                    for s in selection_scenario.selector_performance_data.solvers
                    if s != SelectionScenario.__selector_solver_name__
                ],
                instances=test_set_instances,
                objective=selection_scenario.objective,
            )
            self.test_set_performance[test_set] = test_perf
        self.actual_performance_data = (
            selection_scenario.selector_performance_data.get_value(
                solver=SelectionScenario.__selector_solver_name__,
                instance=self.training_instances,
                objective=self.objective.name,
            )
        )
        self.actual_performance = self.objective.instance_aggregator(
            self.actual_performance_data
        )

    def get_solver_data(
        self: SelectionOutput, train_data: PerformanceDataFrame
    ) -> SelectionSolverData:
        """Initalise SelectionSolverData object."""
        num_solvers = train_data.num_solvers
        return SelectionSolverData(self.solver_performance_ranking, num_solvers)

    def serialise_solvers(self: SelectionOutput, sd: SelectionSolverData) -> dict:
        """Transform SelectionSolverData to dictionary format."""
        return {
            "number_of_solvers": sd.num_solvers,
            "single_best_solver": sd.single_best_solver,
            "solver_ranking": [
                {"solver_name": solver[0], "performance": solver[1]}
                for solver in sd.solver_performance_ranking
            ],
        }

    def serialise_performance(self: SelectionOutput, sp: SelectionPerformance) -> dict:
        """Transform SelectionPerformance to dictionary format."""
        return {
            "vbs_performance": sp.vbs_performance,
            "actual_performance": sp.actual_performance,
            "objective": self.objective.name,
            "metric": sp.metric,
        }

    def serialise_instances(self: SelectionOutput, instances: list[str]) -> dict:
        """Transform Instances to dictionary format."""
        instance_sets = set(Path(instance).parent.name for instance in instances)
        return {
            "number_of_instance_sets": len(instance_sets),
            "instance_sets": [
                {
                    "name": instance_set,
                    "number_of_instances": sum(
                        [1 if instance_set in instance else 0 for instance in instances]
                    ),
                }
                for instance_set in instance_sets
            ],
        }

    def serialise_marginal_contribution(self: SelectionOutput) -> dict:
        """Transform performance ranking to dictionary format."""
        return {
            "marginal_contribution_actual": [
                {
                    "solver_name": ranking[0],
                    "marginal_contribution": ranking[1],
                    "best_performance": ranking[2],
                }
                for ranking in self.marginal_contribution_actual
            ],
            "marginal_contribution_perfect": [
                {
                    "solver_name": ranking[0],
                    "marginal_contribution": ranking[1],
                    "best_performance": ranking[2],
                }
                for ranking in self.marginal_contribution_perfect
            ],
        }

    def serialise(self: SelectionOutput) -> dict:
        """Serialise the selection output."""
        test_data = (
            self.serialise_instances(self.test_instances)
            if self.test_instances
            else None
        )
        return {
            "solvers": self.serialise_solvers(self.solver_data),
            "training_instances": self.serialise_instances(self.training_instances),
            "test_instances": test_data,
            "settings": {"cutoff_time": self.cutoff_time},
            "marginal_contribution": self.serialise_marginal_contribution(),
        }

    def write_output(self: SelectionOutput, output: Path) -> None:
        """Write data into a JSON file."""
        output = output / "configuration.json" if output.is_dir() else output
        with output.open("w") as f:
            json.dump(self.serialise(), f, indent=4)
